from dataclasses import dataclass
from typing import List

from openapi_core.exceptions import OpenAPIError


class MediaTypeFinderError(OpenAPIError):
    """Media type finder error"""


@dataclass
class MediaTypeNotFound(MediaTypeFinderError):
    mimetype: str
    availableMimetypes: List[str]

    def __str__(self) -> str:
        return (
            f"Content for the following mimetype not found: {self.mimetype}. "
            f"Valid mimetypes: {self.availableMimetypes}"
        )



"""OpenAPI core unmarshalling processors module"""

from openapi_core.typing import RequestType
from openapi_core.typing import ResponseType
from openapi_core.unmarshalling.integrations import (
    AsyncUnmarshallingIntegration,
)
from openapi_core.unmarshalling.integrations import UnmarshallingIntegration
from openapi_core.unmarshalling.typing import AsyncValidRequestHandlerCallable
from openapi_core.unmarshalling.typing import ErrorsHandlerCallable
from openapi_core.unmarshalling.typing import ValidRequestHandlerCallable


class UnmarshallingProcessor(
    UnmarshallingIntegration[RequestType, ResponseType]
):
    def handle_request(
        self,
        request: RequestType,
        valid_handler: ValidRequestHandlerCallable[ResponseType],
        errors_handler: ErrorsHandlerCallable[ResponseType],
    ) -> ResponseType:
        request_unmarshal_result = self.unmarshal_request(
            request,
        )
        if request_unmarshal_result.errors:
            return errors_handler(request_unmarshal_result.errors)
        return valid_handler(request_unmarshal_result)

    def handle_response(
        self,
        request: RequestType,
        response: ResponseType,
        errors_handler: ErrorsHandlerCallable[ResponseType],
    ) -> ResponseType:
        response_unmarshal_result = self.unmarshal_response(request, response)
        if response_unmarshal_result.errors:
            return errors_handler(response_unmarshal_result.errors)
        return response


class AsyncUnmarshallingProcessor(
    AsyncUnmarshallingIntegration[RequestType, ResponseType]
):
    async def handle_request(
        self,
        request: RequestType,
        valid_handler: AsyncValidRequestHandlerCallable[ResponseType],
        errors_handler: ErrorsHandlerCallable[ResponseType],
    ) -> ResponseType:
        request_unmarshal_result = await self.unmarshal_request(request)
        if request_unmarshal_result.errors:
            return errors_handler(request_unmarshal_result.errors)
        result = await valid_handler(request_unmarshal_result)
        return result

    async def handle_response(
        self,
        request: RequestType,
        response: ResponseType,
        errors_handler: ErrorsHandlerCallable[ResponseType],
    ) -> ResponseType:
        response_unmarshal_result = await self.unmarshal_response(
            request, response
        )
        if response_unmarshal_result.errors:
            return errors_handler(response_unmarshal_result.errors)
        return response



from dataclasses import dataclass

from openapi_core.exceptions import OpenAPIError


class UnmarshalError(OpenAPIError):
    """Schema unmarshal operation error"""


class UnmarshallerError(UnmarshalError):
    """Unmarshaller error"""


@dataclass
class FormatterNotFoundError(UnmarshallerError):
    """Formatter not found to unmarshal"""

    type_format: str

    def __str__(self) -> str:
        return f"Formatter not found for {self.type_format} format"



"""OpenAPI core contrib requests requests module"""

from typing import Optional
from typing import Union
from urllib.parse import parse_qs
from urllib.parse import urlparse

from requests import PreparedRequest
from requests import Request
from requests.cookies import RequestsCookieJar
from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict

from openapi_core.contrib.requests.protocols import SupportsCookieJar
from openapi_core.datatypes import RequestParameters


class RequestsOpenAPIRequest:
    """
    Converts a requests request to an OpenAPI request

    Internally converts to a `PreparedRequest` first to parse the exact
    payload being sent
    """

    def __init__(self, request: Union[Request, PreparedRequest]):
        if not isinstance(request, (Request, PreparedRequest)):
            raise TypeError(
                "'request' argument is not type of "
                f"{Request} or {PreparedRequest}"
            )
        if isinstance(request, Request):
            request = request.prepare()

        self.request = request
        if request.url is None:
            raise RuntimeError("Request URL is missing")
        self._url_parsed = urlparse(request.url, allow_fragments=False)

        cookie = {}
        if isinstance(self.request, SupportsCookieJar) and isinstance(
            self.request._cookies, RequestsCookieJar
        ):
            # cookies are stored in a cookiejar object
            cookie = self.request._cookies.get_dict()

        self.parameters = RequestParameters(
            query=ImmutableMultiDict(parse_qs(self._url_parsed.query)),
            header=Headers(dict(self.request.headers)),
            cookie=ImmutableMultiDict(cookie),
        )

    @property
    def host_url(self) -> str:
        return f"{self._url_parsed.scheme}://{self._url_parsed.netloc}"

    @property
    def path(self) -> str:
        assert isinstance(self._url_parsed.path, str)
        return self._url_parsed.path

    @property
    def method(self) -> str:
        method = self.request.method
        return method and method.lower() or ""

    @property
    def body(self) -> Optional[bytes]:
        if self.request.body is None:
            return None
        if isinstance(self.request.body, bytes):
            return self.request.body
        assert isinstance(self.request.body, str)
        # TODO: figure out if request._body_position is relevant
        return self.request.body.encode("utf-8")

    @property
    def content_type(self) -> str:
        # Order matters because all python requests issued from a session
        # include Accept */* which does not necessarily match the content type
        return str(
            self.request.headers.get("Content-Type")
            or self.request.headers.get("Accept")
        )


class RequestsOpenAPIWebhookRequest(RequestsOpenAPIRequest):
    """
    Converts a requests request to an OpenAPI Webhook request

    Internally converts to a `PreparedRequest` first to parse the exact
    payload being sent
    """

    def __init__(self, request: Union[Request, PreparedRequest], name: str):
        super().__init__(request)
        self.name = name



from typing import Type
from typing import Union

from openapi_core.unmarshalling.request.protocols import RequestUnmarshaller
from openapi_core.unmarshalling.request.protocols import (
    WebhookRequestUnmarshaller,
)

RequestUnmarshallerType = Type[RequestUnmarshaller]
WebhookRequestUnmarshallerType = Type[WebhookRequestUnmarshaller]
AnyRequestUnmarshallerType = Union[
    RequestUnmarshallerType, WebhookRequestUnmarshallerType
]



from dataclasses import dataclass
from typing import List

from openapi_core.exceptions import OpenAPIError


class SecurityFinderError(OpenAPIError):
    """Security finder error"""


@dataclass
class SecurityNotFound(SecurityFinderError):
    """Find security error"""

    schemes: List[List[str]]

    def __str__(self) -> str:
        return f"Security not found. Schemes not valid for any requirement: {str(self.schemes)}"



from openapi_schema_validator import OAS30ReadValidator
from openapi_schema_validator import OAS30WriteValidator
from openapi_schema_validator import OAS31Validator

from openapi_core.validation.schemas.factories import SchemaValidatorsFactory

__all__ = [
    "oas30_write_schema_validators_factory",
    "oas30_read_schema_validators_factory",
    "oas31_schema_validators_factory",
]

oas30_write_schema_validators_factory = SchemaValidatorsFactory(
    OAS30WriteValidator,
)

oas30_read_schema_validators_factory = SchemaValidatorsFactory(
    OAS30ReadValidator,
)

oas31_schema_validators_factory = SchemaValidatorsFactory(
    OAS31Validator,
    # FIXME: OpenAPI 3.1 schema validator uses OpenAPI 3.0 format checker.
    # See https://github.com/python-openapi/openapi-core/issues/506
    format_checker=OAS30ReadValidator.FORMAT_CHECKER,
)



from dataclasses import dataclass
from typing import List

from openapi_core.exceptions import OpenAPIError


class ResponseFinderError(OpenAPIError):
    """Response finder error"""


@dataclass
class ResponseNotFound(ResponseFinderError):
    """Find response error"""

    http_status: str
    availableresponses: List[str]

    def __str__(self) -> str:
        return f"Unknown response http status: {str(self.http_status)}"



"""OpenAPI core testing responses module"""

from typing import Any
from typing import Dict
from typing import Optional

from werkzeug.datastructures import Headers


class MockResponse:
    def __init__(
        self,
        data: bytes,
        status_code: int = 200,
        headers: Optional[Dict[str, Any]] = None,
        content_type: str = "application/json",
    ):
        self.data = data
        self.status_code = status_code
        self.headers = Headers(headers or {})
        self.content_type = content_type



from typing import Any
from typing import Callable
from typing import Dict
from typing import Mapping

DeserializerCallable = Callable[[bool, str, str, Mapping[str, Any]], Any]
StyleDeserializersDict = Dict[str, DeserializerCallable]



from dataclasses import dataclass
from typing import Optional

from openapi_core.unmarshalling.schemas.datatypes import (
    FormatUnmarshallersDict,
)
from openapi_core.unmarshalling.schemas.factories import (
    SchemaUnmarshallersFactory,
)
from openapi_core.validation.configurations import ValidatorConfig


@dataclass
class UnmarshallerConfig(ValidatorConfig):
    """Unmarshaller configuration dataclass.

    Attributes:
        schema_unmarshallers_factory
            Schema unmarshallers factory.
        extra_format_unmarshallers
            Extra format unmarshallers.
    """

    schema_unmarshallers_factory: Optional[SchemaUnmarshallersFactory] = None
    extra_format_unmarshallers: Optional[FormatUnmarshallersDict] = None



import warnings
from typing import Optional

from jsonschema_path import SchemaPath

from openapi_core.unmarshalling.schemas.datatypes import (
    FormatUnmarshallersDict,
)
from openapi_core.unmarshalling.schemas.exceptions import (
    FormatterNotFoundError,
)
from openapi_core.unmarshalling.schemas.unmarshallers import (
    FormatsUnmarshaller,
)
from openapi_core.unmarshalling.schemas.unmarshallers import SchemaUnmarshaller
from openapi_core.unmarshalling.schemas.unmarshallers import TypesUnmarshaller
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory


class SchemaUnmarshallersFactory:
    def __init__(
        self,
        schema_validators_factory: SchemaValidatorsFactory,
        types_unmarshaller: TypesUnmarshaller,
        format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
    ):
        self.schema_validators_factory = schema_validators_factory
        self.types_unmarshaller = types_unmarshaller
        if format_unmarshallers is None:
            format_unmarshallers = {}
        self.format_unmarshallers = format_unmarshallers

    def create(
        self,
        schema: SchemaPath,
        format_validators: Optional[FormatValidatorsDict] = None,
        format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
    ) -> SchemaUnmarshaller:
        """Create unmarshaller from the schema."""
        if schema is None:
            raise TypeError("Invalid schema")

        if schema.getkey("deprecated", False):
            warnings.warn("The schema is deprecated", DeprecationWarning)

        if extra_format_validators is None:
            extra_format_validators = {}
        schema_validator = self.schema_validators_factory.create(
            schema,
            format_validators=format_validators,
            extra_format_validators=extra_format_validators,
        )

        schema_format = schema.getkey("format")

        formats_unmarshaller = FormatsUnmarshaller(
            format_unmarshallers or self.format_unmarshallers,
            extra_format_unmarshallers,
        )

        # FIXME: don;t raise exception on unknown format
        # See https://github.com/python-openapi/openapi-core/issues/515
        if (
            schema_format
            and schema_format not in schema_validator
            and schema_format not in formats_unmarshaller
        ):
            raise FormatterNotFoundError(schema_format)

        return SchemaUnmarshaller(
            schema,
            schema_validator,
            self.types_unmarshaller,
            formats_unmarshaller,
        )



"""OpenAPI core validation request protocols module"""

from typing import Optional
from typing import Protocol
from typing import runtime_checkable

from jsonschema_path import SchemaPath
from openapi_spec_validator.validation.types import SpecValidatorType

from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.deserializing.media_types import (
    media_type_deserializers_factory,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.styles import style_deserializers_factory
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.protocols import Request
from openapi_core.protocols import WebhookRequest
from openapi_core.security import security_provider_factory
from openapi_core.security.factories import SecurityProviderFactory
from openapi_core.templating.paths.types import PathFinderType
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.schemas.datatypes import (
    FormatUnmarshallersDict,
)
from openapi_core.unmarshalling.schemas.factories import (
    SchemaUnmarshallersFactory,
)
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory


@runtime_checkable
class RequestUnmarshaller(Protocol):
    def __init__(
        self,
        spec: SchemaPath,
        base_url: Optional[str] = None,
        style_deserializers_factory: StyleDeserializersFactory = style_deserializers_factory,
        media_type_deserializers_factory: MediaTypeDeserializersFactory = media_type_deserializers_factory,
        schema_casters_factory: Optional[SchemaCastersFactory] = None,
        schema_validators_factory: Optional[SchemaValidatorsFactory] = None,
        path_finder_cls: Optional[PathFinderType] = None,
        spec_validator_cls: Optional[SpecValidatorType] = None,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
        security_provider_factory: SecurityProviderFactory = security_provider_factory,
        schema_unmarshallers_factory: Optional[
            SchemaUnmarshallersFactory
        ] = None,
        format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
        extra_format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
    ): ...

    def unmarshal(
        self,
        request: Request,
    ) -> RequestUnmarshalResult: ...


@runtime_checkable
class WebhookRequestUnmarshaller(Protocol):
    def __init__(
        self,
        spec: SchemaPath,
        base_url: Optional[str] = None,
        style_deserializers_factory: StyleDeserializersFactory = style_deserializers_factory,
        media_type_deserializers_factory: MediaTypeDeserializersFactory = media_type_deserializers_factory,
        schema_casters_factory: Optional[SchemaCastersFactory] = None,
        schema_validators_factory: Optional[SchemaValidatorsFactory] = None,
        path_finder_cls: Optional[PathFinderType] = None,
        spec_validator_cls: Optional[SpecValidatorType] = None,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
        security_provider_factory: SecurityProviderFactory = security_provider_factory,
        schema_unmarshallers_factory: Optional[
            SchemaUnmarshallersFactory
        ] = None,
        format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
        extra_format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
    ): ...

    def unmarshal(
        self,
        request: WebhookRequest,
    ) -> RequestUnmarshalResult: ...



import re
from functools import partial
from typing import Any
from typing import List
from typing import Mapping

from openapi_core.schema.protocols import SuportsGetAll
from openapi_core.schema.protocols import SuportsGetList


def split(value: str, separator: str = ",", step: int = 1) -> List[str]:
    parts = value.split(separator)

    if step == 1:
        return parts

    result = []
    for i in range(len(parts)):
        if i % step == 0:
            if i + 1 < len(parts):
                result.append(parts[i] + separator + parts[i + 1])
    return result


def delimited_loads(
    explode: bool,
    name: str,
    schema_type: str,
    location: Mapping[str, Any],
    delimiter: str,
) -> Any:
    value = location[name]

    explode_type = (explode, schema_type)
    if explode_type == (False, "array"):
        return split(value, separator=delimiter)
    if explode_type == (False, "object"):
        return dict(
            map(
                partial(split, separator=delimiter),
                split(value, separator=delimiter, step=2),
            )
        )

    raise ValueError("not available")


def matrix_loads(
    explode: bool, name: str, schema_type: str, location: Mapping[str, Any]
) -> Any:
    if explode == False:
        m = re.match(rf"^;{name}=(.*)$", location[f";{name}"])
        if m is None:
            raise KeyError(name)
        value = m.group(1)
        # ;color=blue,black,brown
        if schema_type == "array":
            return split(value)
        # ;color=R,100,G,200,B,150
        if schema_type == "object":
            return dict(map(split, split(value, step=2)))
        # .;color=blue
        return value
    else:
        # ;color=blue;color=black;color=brown
        if schema_type == "array":
            return re.findall(rf";{name}=([^;]*)", location[f";{name}*"])
        # ;R=100;G=200;B=150
        if schema_type == "object":
            value = location[f";{name}*"]
            return dict(
                map(
                    partial(split, separator="="),
                    split(value[1:], separator=";"),
                )
            )
        # ;color=blue
        m = re.match(rf"^;{name}=(.*)$", location[f";{name}*"])
        if m is None:
            raise KeyError(name)
        value = m.group(1)
        return value


def label_loads(
    explode: bool, name: str, schema_type: str, location: Mapping[str, Any]
) -> Any:
    if explode == False:
        value = location[f".{name}"]
        # .blue,black,brown
        if schema_type == "array":
            return split(value[1:])
        # .R,100,G,200,B,150
        if schema_type == "object":
            return dict(map(split, split(value[1:], separator=",", step=2)))
        # .blue
        return value[1:]
    else:
        value = location[f".{name}*"]
        # .blue.black.brown
        if schema_type == "array":
            return split(value[1:], separator=".")
        # .R=100.G=200.B=150
        if schema_type == "object":
            return dict(
                map(
                    partial(split, separator="="),
                    split(value[1:], separator="."),
                )
            )
        # .blue
        return value[1:]


def form_loads(
    explode: bool, name: str, schema_type: str, location: Mapping[str, Any]
) -> Any:
    explode_type = (explode, schema_type)
    # color=blue,black,brown
    if explode_type == (False, "array"):
        return split(location[name], separator=",")
    # color=blue&color=black&color=brown
    elif explode_type == (True, "array"):
        if name not in location:
            raise KeyError(name)
        if isinstance(location, SuportsGetAll):
            return location.getall(name)
        if isinstance(location, SuportsGetList):
            return location.getlist(name)
        return location[name]

    value = location[name]
    # color=R,100,G,200,B,150
    if explode_type == (False, "object"):
        return dict(map(split, split(value, separator=",", step=2)))
    # R=100&G=200&B=150
    elif explode_type == (True, "object"):
        return dict(
            map(partial(split, separator="="), split(value, separator="&"))
        )

    # color=blue
    return value


def simple_loads(
    explode: bool, name: str, schema_type: str, location: Mapping[str, Any]
) -> Any:
    value = location[name]

    # blue,black,brown
    if schema_type == "array":
        return split(value, separator=",")

    explode_type = (explode, schema_type)
    # R,100,G,200,B,150
    if explode_type == (False, "object"):
        return dict(map(split, split(value, separator=",", step=2)))
    # R=100,G=200,B=150
    elif explode_type == (True, "object"):
        return dict(
            map(partial(split, separator="="), split(value, separator=","))
        )

    # blue
    return value


def space_delimited_loads(
    explode: bool, name: str, schema_type: str, location: Mapping[str, Any]
) -> Any:
    return delimited_loads(
        explode, name, schema_type, location, delimiter="%20"
    )


def pipe_delimited_loads(
    explode: bool, name: str, schema_type: str, location: Mapping[str, Any]
) -> Any:
    return delimited_loads(explode, name, schema_type, location, delimiter="|")


def deep_object_loads(
    explode: bool, name: str, schema_type: str, location: Mapping[str, Any]
) -> Any:
    explode_type = (explode, schema_type)

    if explode_type != (True, "object"):
        raise ValueError("not available")

    keys_str = " ".join(location.keys())
    if not re.search(rf"{name}\[\w+\]", keys_str):
        raise KeyError(name)

    values = {}
    for key, value in location.items():
        # Split the key from the brackets.
        key_split = re.split(pattern=r"\[|\]", string=key)
        if key_split[0] == name:
            values[key_split[1]] = value
    return values



from jsonschema_path import SchemaPath

from openapi_core.templating.responses.exceptions import ResponseNotFound


class ResponseFinder:
    def __init__(self, responses: SchemaPath):
        self.responses = responses

    def find(self, http_status: str = "default") -> SchemaPath:
        if http_status in self.responses:
            return self.responses / http_status

        # try range
        http_status_range = f"{http_status[0]}XX"
        if http_status_range in self.responses:
            return self.responses / http_status_range

        if "default" not in self.responses:
            raise ResponseNotFound(http_status, list(self.responses.keys()))

        return self.responses / "default"



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Iterable

from openapi_core.exceptions import OpenAPIError
from openapi_core.validation.exceptions import ValidationError
from openapi_core.validation.schemas.exceptions import ValidateError


@dataclass
class HeadersError(Exception):
    headers: Dict[str, Any]
    context: Iterable[OpenAPIError]


class ResponseValidationError(ValidationError):
    """Response error"""


class DataValidationError(ResponseValidationError):
    """Data error"""


class InvalidData(DataValidationError, ValidateError):
    """Invalid data"""


class MissingData(DataValidationError):
    def __str__(self) -> str:
        return "Missing response data"


@dataclass
class HeaderValidationError(ResponseValidationError):
    name: str


class InvalidHeader(HeaderValidationError, ValidateError):
    """Invalid header"""


class MissingHeaderError(HeaderValidationError):
    """Missing header error"""


class MissingHeader(MissingHeaderError):
    def __str__(self) -> str:
        return f"Missing header (without default value): {self.name}"


class MissingRequiredHeader(MissingHeaderError):
    def __str__(self) -> str:
        return f"Missing required header: {self.name}"



"""OpenAPI core module"""

from openapi_core.app import OpenAPI
from openapi_core.configurations import Config
from openapi_core.shortcuts import unmarshal_apicall_request
from openapi_core.shortcuts import unmarshal_apicall_response
from openapi_core.shortcuts import unmarshal_request
from openapi_core.shortcuts import unmarshal_response
from openapi_core.shortcuts import unmarshal_webhook_request
from openapi_core.shortcuts import unmarshal_webhook_response
from openapi_core.shortcuts import validate_apicall_request
from openapi_core.shortcuts import validate_apicall_response
from openapi_core.shortcuts import validate_request
from openapi_core.shortcuts import validate_response
from openapi_core.shortcuts import validate_webhook_request
from openapi_core.shortcuts import validate_webhook_response
from openapi_core.spec.paths import Spec
from openapi_core.unmarshalling.request import V3RequestUnmarshaller
from openapi_core.unmarshalling.request import V3WebhookRequestUnmarshaller
from openapi_core.unmarshalling.request import V30RequestUnmarshaller
from openapi_core.unmarshalling.request import V31RequestUnmarshaller
from openapi_core.unmarshalling.request import V31WebhookRequestUnmarshaller
from openapi_core.unmarshalling.response import V3ResponseUnmarshaller
from openapi_core.unmarshalling.response import V3WebhookResponseUnmarshaller
from openapi_core.unmarshalling.response import V30ResponseUnmarshaller
from openapi_core.unmarshalling.response import V31ResponseUnmarshaller
from openapi_core.unmarshalling.response import V31WebhookResponseUnmarshaller
from openapi_core.validation.request import V3RequestValidator
from openapi_core.validation.request import V3WebhookRequestValidator
from openapi_core.validation.request import V30RequestValidator
from openapi_core.validation.request import V31RequestValidator
from openapi_core.validation.request import V31WebhookRequestValidator
from openapi_core.validation.response import V3ResponseValidator
from openapi_core.validation.response import V3WebhookResponseValidator
from openapi_core.validation.response import V30ResponseValidator
from openapi_core.validation.response import V31ResponseValidator
from openapi_core.validation.response import V31WebhookResponseValidator

__author__ = "Artur Maciag"
__email__ = "maciag.artur@gmail.com"
__version__ = "0.19.5"
__url__ = "https://github.com/python-openapi/openapi-core"
__license__ = "BSD 3-Clause License"

__all__ = [
    "OpenAPI",
    "Config",
    "Spec",
    "unmarshal_request",
    "unmarshal_response",
    "unmarshal_apicall_request",
    "unmarshal_webhook_request",
    "unmarshal_apicall_response",
    "unmarshal_webhook_response",
    "validate_apicall_request",
    "validate_webhook_request",
    "validate_apicall_response",
    "validate_webhook_response",
    "validate_request",
    "validate_response",
    "V30RequestUnmarshaller",
    "V30ResponseUnmarshaller",
    "V31RequestUnmarshaller",
    "V31ResponseUnmarshaller",
    "V31WebhookRequestUnmarshaller",
    "V31WebhookResponseUnmarshaller",
    "V3RequestUnmarshaller",
    "V3ResponseUnmarshaller",
    "V3WebhookRequestUnmarshaller",
    "V3WebhookResponseUnmarshaller",
    "V30RequestValidator",
    "V30ResponseValidator",
    "V31RequestValidator",
    "V31ResponseValidator",
    "V31WebhookRequestValidator",
    "V31WebhookResponseValidator",
    "V3RequestValidator",
    "V3ResponseValidator",
    "V3WebhookRequestValidator",
    "V3WebhookResponseValidator",
]



"""OpenAPI core unmarshalling request datatypes module"""

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import Any

from openapi_core.datatypes import Parameters
from openapi_core.unmarshalling.datatypes import BaseUnmarshalResult


@dataclass
class RequestUnmarshalResult(BaseUnmarshalResult):
    body: Any | None = None
    parameters: Parameters = field(default_factory=Parameters)
    security: dict[str, str] | None = None



from dataclasses import dataclass
from typing import Any

from openapi_core.exceptions import OpenAPIError


@dataclass
class CastError(OpenAPIError):
    """Schema cast operation error"""

    value: Any
    type: str

    def __str__(self) -> str:
        return f"Failed to cast value to {self.type} type: {self.value}"



"""OpenAPI core validation request validators module"""

import warnings
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Optional

from jsonschema_path import SchemaPath
from openapi_spec_validator import OpenAPIV30SpecValidator
from openapi_spec_validator import OpenAPIV31SpecValidator
from openapi_spec_validator.validation.types import SpecValidatorType

from openapi_core.casting.schemas import oas30_write_schema_casters_factory
from openapi_core.casting.schemas import oas31_schema_casters_factory
from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.datatypes import Parameters
from openapi_core.datatypes import RequestParameters
from openapi_core.deserializing.media_types import (
    media_type_deserializers_factory,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.styles import style_deserializers_factory
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.protocols import BaseRequest
from openapi_core.protocols import Request
from openapi_core.protocols import WebhookRequest
from openapi_core.security import security_provider_factory
from openapi_core.security.exceptions import SecurityProviderError
from openapi_core.security.factories import SecurityProviderFactory
from openapi_core.templating.paths.exceptions import PathError
from openapi_core.templating.paths.types import PathFinderType
from openapi_core.templating.security.exceptions import SecurityNotFound
from openapi_core.util import chainiters
from openapi_core.validation.decorators import ValidationErrorWrapper
from openapi_core.validation.request.exceptions import InvalidParameter
from openapi_core.validation.request.exceptions import InvalidRequestBody
from openapi_core.validation.request.exceptions import InvalidSecurity
from openapi_core.validation.request.exceptions import MissingParameter
from openapi_core.validation.request.exceptions import MissingRequestBody
from openapi_core.validation.request.exceptions import MissingRequiredParameter
from openapi_core.validation.request.exceptions import (
    MissingRequiredRequestBody,
)
from openapi_core.validation.request.exceptions import ParametersError
from openapi_core.validation.request.exceptions import ParameterValidationError
from openapi_core.validation.request.exceptions import (
    RequestBodyValidationError,
)
from openapi_core.validation.request.exceptions import SecurityValidationError
from openapi_core.validation.schemas import (
    oas30_write_schema_validators_factory,
)
from openapi_core.validation.schemas import oas31_schema_validators_factory
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory
from openapi_core.validation.validators import BaseAPICallValidator
from openapi_core.validation.validators import BaseValidator
from openapi_core.validation.validators import BaseWebhookValidator


class BaseRequestValidator(BaseValidator):
    def __init__(
        self,
        spec: SchemaPath,
        base_url: Optional[str] = None,
        style_deserializers_factory: StyleDeserializersFactory = style_deserializers_factory,
        media_type_deserializers_factory: MediaTypeDeserializersFactory = media_type_deserializers_factory,
        schema_casters_factory: Optional[SchemaCastersFactory] = None,
        schema_validators_factory: Optional[SchemaValidatorsFactory] = None,
        path_finder_cls: Optional[PathFinderType] = None,
        spec_validator_cls: Optional[SpecValidatorType] = None,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
        security_provider_factory: SecurityProviderFactory = security_provider_factory,
    ):

        BaseValidator.__init__(
            self,
            spec,
            base_url=base_url,
            style_deserializers_factory=style_deserializers_factory,
            media_type_deserializers_factory=media_type_deserializers_factory,
            schema_casters_factory=schema_casters_factory,
            schema_validators_factory=schema_validators_factory,
            path_finder_cls=path_finder_cls,
            spec_validator_cls=spec_validator_cls,
            format_validators=format_validators,
            extra_format_validators=extra_format_validators,
            extra_media_type_deserializers=extra_media_type_deserializers,
        )
        self.security_provider_factory = security_provider_factory

    def _iter_errors(
        self, request: BaseRequest, operation: SchemaPath, path: SchemaPath
    ) -> Iterator[Exception]:
        try:
            self._get_security(request.parameters, operation)
        # don't process if security errors
        except SecurityValidationError as exc:
            yield exc
            return

        try:
            self._get_parameters(request.parameters, operation, path)
        except ParametersError as exc:
            yield from exc.errors

        try:
            self._get_body(request.body, request.content_type, operation)
        except MissingRequestBody:
            pass
        except RequestBodyValidationError as exc:
            yield exc

    def _iter_body_errors(
        self, request: BaseRequest, operation: SchemaPath
    ) -> Iterator[Exception]:
        try:
            self._get_body(request.body, request.content_type, operation)
        except RequestBodyValidationError as exc:
            yield exc

    def _iter_parameters_errors(
        self, request: BaseRequest, operation: SchemaPath, path: SchemaPath
    ) -> Iterator[Exception]:
        try:
            self._get_parameters(request.parameters, path, operation)
        except ParametersError as exc:
            yield from exc.errors

    def _iter_security_errors(
        self, request: BaseRequest, operation: SchemaPath
    ) -> Iterator[Exception]:
        try:
            self._get_security(request.parameters, operation)
        except SecurityValidationError as exc:
            yield exc

    def _get_parameters(
        self,
        parameters: RequestParameters,
        operation: SchemaPath,
        path: SchemaPath,
    ) -> Parameters:
        operation_params = operation.get("parameters", [])
        path_params = path.get("parameters", [])

        errors = []
        seen = set()
        validated = Parameters()
        params_iter = chainiters(operation_params, path_params)
        for param in params_iter:
            param_name = param["name"]
            param_location = param["in"]
            if (param_name, param_location) in seen:
                # skip parameter already seen
                # e.g. overriden path item paremeter on operation
                continue
            seen.add((param_name, param_location))
            try:
                value = self._get_parameter(parameters, param)
            except MissingParameter:
                continue
            except ParameterValidationError as exc:
                errors.append(exc)
                continue
            else:
                location = getattr(validated, param_location)
                location[param_name] = value

        if errors:
            raise ParametersError(errors=errors, parameters=validated)

        return validated

    @ValidationErrorWrapper(
        ParameterValidationError,
        InvalidParameter,
        "from_spec",
        spec="param",
    )
    def _get_parameter(
        self, parameters: RequestParameters, param: SchemaPath
    ) -> Any:
        name = param["name"]
        deprecated = param.getkey("deprecated", False)
        if deprecated:
            warnings.warn(
                f"{name} parameter is deprecated",
                DeprecationWarning,
            )

        param_location = param["in"]
        location = parameters[param_location]

        try:
            value, _ = self._get_param_or_header_and_schema(param, location)
        except KeyError:
            required = param.getkey("required", False)
            if required:
                raise MissingRequiredParameter(name, param_location)
            raise MissingParameter(name, param_location)
        else:
            return value

    @ValidationErrorWrapper(SecurityValidationError, InvalidSecurity)
    def _get_security(
        self, parameters: RequestParameters, operation: SchemaPath
    ) -> Optional[Dict[str, str]]:
        security = None
        if "security" in self.spec:
            security = self.spec / "security"
        if "security" in operation:
            security = operation / "security"

        if not security:
            return {}

        schemes = []
        for security_requirement in security:
            try:
                scheme_names = list(security_requirement.keys())
                schemes.append(scheme_names)
                return {
                    scheme_name: self._get_security_value(
                        parameters, scheme_name
                    )
                    for scheme_name in scheme_names
                }
            except SecurityProviderError:
                continue

        raise SecurityNotFound(schemes)

    def _get_security_value(
        self, parameters: RequestParameters, scheme_name: str
    ) -> Any:
        security_schemes = self.spec / "components#securitySchemes"
        if scheme_name not in security_schemes:
            return
        scheme = security_schemes[scheme_name]
        security_provider = self.security_provider_factory.create(scheme)
        return security_provider(parameters)

    @ValidationErrorWrapper(RequestBodyValidationError, InvalidRequestBody)
    def _get_body(
        self, body: Optional[bytes], mimetype: str, operation: SchemaPath
    ) -> Any:
        if "requestBody" not in operation:
            return None

        # TODO: implement required flag checking
        request_body = operation / "requestBody"
        content = request_body / "content"

        raw_body = self._get_body_value(body, request_body)
        value, _ = self._get_content_and_schema(raw_body, content, mimetype)
        return value

    def _get_body_value(
        self, body: Optional[bytes], request_body: SchemaPath
    ) -> bytes:
        if not body:
            if request_body.getkey("required", False):
                raise MissingRequiredRequestBody
            raise MissingRequestBody
        return body


class BaseAPICallRequestValidator(BaseRequestValidator, BaseAPICallValidator):
    def iter_errors(self, request: Request) -> Iterator[Exception]:
        raise NotImplementedError

    def validate(self, request: Request) -> None:
        for err in self.iter_errors(request):
            raise err


class BaseWebhookRequestValidator(BaseRequestValidator, BaseWebhookValidator):
    def iter_errors(self, request: WebhookRequest) -> Iterator[Exception]:
        raise NotImplementedError

    def validate(self, request: WebhookRequest) -> None:
        for err in self.iter_errors(request):
            raise err


class APICallRequestBodyValidator(BaseAPICallRequestValidator):
    def iter_errors(self, request: Request) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        except PathError as exc:
            yield exc
            return

        yield from self._iter_body_errors(request, operation)


class APICallRequestParametersValidator(BaseAPICallRequestValidator):
    def iter_errors(self, request: Request) -> Iterator[Exception]:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        except PathError as exc:
            yield exc
            return

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        yield from self._iter_parameters_errors(request, operation, path)


class APICallRequestSecurityValidator(BaseAPICallRequestValidator):
    def iter_errors(self, request: Request) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        except PathError as exc:
            yield exc
            return

        yield from self._iter_security_errors(request, operation)


class APICallRequestValidator(BaseAPICallRequestValidator):
    def iter_errors(self, request: Request) -> Iterator[Exception]:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        yield from self._iter_errors(request, operation, path)


class WebhookRequestValidator(BaseWebhookRequestValidator):
    def iter_errors(self, request: WebhookRequest) -> Iterator[Exception]:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        yield from self._iter_errors(request, operation, path)


class WebhookRequestBodyValidator(BaseWebhookRequestValidator):
    def iter_errors(self, request: WebhookRequest) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        except PathError as exc:
            yield exc
            return

        yield from self._iter_body_errors(request, operation)


class WebhookRequestParametersValidator(BaseWebhookRequestValidator):
    def iter_errors(self, request: WebhookRequest) -> Iterator[Exception]:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        except PathError as exc:
            yield exc
            return

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        yield from self._iter_parameters_errors(request, operation, path)


class WebhookRequestSecurityValidator(BaseWebhookRequestValidator):
    def iter_errors(self, request: WebhookRequest) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        except PathError as exc:
            yield exc
            return

        yield from self._iter_security_errors(request, operation)


class V30RequestBodyValidator(APICallRequestBodyValidator):
    spec_validator_cls = OpenAPIV30SpecValidator
    schema_casters_factory = oas30_write_schema_casters_factory
    schema_validators_factory = oas30_write_schema_validators_factory


class V30RequestParametersValidator(APICallRequestParametersValidator):
    spec_validator_cls = OpenAPIV30SpecValidator
    schema_casters_factory = oas30_write_schema_casters_factory
    schema_validators_factory = oas30_write_schema_validators_factory


class V30RequestSecurityValidator(APICallRequestSecurityValidator):
    spec_validator_cls = OpenAPIV30SpecValidator
    schema_casters_factory = oas30_write_schema_casters_factory
    schema_validators_factory = oas30_write_schema_validators_factory


class V30RequestValidator(APICallRequestValidator):
    spec_validator_cls = OpenAPIV30SpecValidator
    schema_casters_factory = oas30_write_schema_casters_factory
    schema_validators_factory = oas30_write_schema_validators_factory


class V31RequestBodyValidator(APICallRequestBodyValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31RequestParametersValidator(APICallRequestParametersValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31RequestSecurityValidator(APICallRequestSecurityValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31RequestValidator(APICallRequestValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookRequestBodyValidator(WebhookRequestBodyValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookRequestParametersValidator(WebhookRequestParametersValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookRequestSecurityValidator(WebhookRequestSecurityValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookRequestValidator(WebhookRequestValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory



"""OpenAPI core contrib aiohttp responses module"""

import multidict
from aiohttp import web


class AIOHTTPOpenAPIWebResponse:
    def __init__(self, response: web.Response):
        if not isinstance(response, web.Response):
            raise TypeError(
                f"'response' argument is not type of {web.Response.__qualname__!r}"
            )
        self.response = response

    @property
    def data(self) -> bytes:
        if self.response.body is None:
            return b""
        if isinstance(self.response.body, bytes):
            return self.response.body
        assert isinstance(self.response.body, str)
        return self.response.body.encode("utf-8")

    @property
    def status_code(self) -> int:
        return self.response.status

    @property
    def content_type(self) -> str:
        return self.response.content_type or ""

    @property
    def headers(self) -> multidict.CIMultiDict[str]:
        return self.response.headers



from jsonschema_path import SchemaPath

from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.protocols import WebhookRequest
from openapi_core.templating.paths.exceptions import PathError
from openapi_core.templating.responses.exceptions import ResponseFinderError
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.unmarshalling.schemas import (
    oas30_read_schema_unmarshallers_factory,
)
from openapi_core.unmarshalling.schemas import (
    oas31_schema_unmarshallers_factory,
)
from openapi_core.unmarshalling.unmarshallers import BaseUnmarshaller
from openapi_core.util import chainiters
from openapi_core.validation.response.exceptions import DataValidationError
from openapi_core.validation.response.exceptions import HeadersError
from openapi_core.validation.response.validators import (
    APICallResponseValidator,
)
from openapi_core.validation.response.validators import BaseResponseValidator
from openapi_core.validation.response.validators import (
    V30ResponseDataValidator,
)
from openapi_core.validation.response.validators import (
    V30ResponseHeadersValidator,
)
from openapi_core.validation.response.validators import V30ResponseValidator
from openapi_core.validation.response.validators import (
    V31ResponseDataValidator,
)
from openapi_core.validation.response.validators import (
    V31ResponseHeadersValidator,
)
from openapi_core.validation.response.validators import V31ResponseValidator
from openapi_core.validation.response.validators import (
    V31WebhookResponseDataValidator,
)
from openapi_core.validation.response.validators import (
    V31WebhookResponseHeadersValidator,
)
from openapi_core.validation.response.validators import (
    V31WebhookResponseValidator,
)
from openapi_core.validation.response.validators import (
    WebhookResponseValidator,
)


class BaseResponseUnmarshaller(BaseResponseValidator, BaseUnmarshaller):
    def _unmarshal(
        self,
        response: Response,
        operation: SchemaPath,
    ) -> ResponseUnmarshalResult:
        try:
            operation_response = self._find_operation_response(
                response.status_code, operation
            )
        # don't process if operation errors
        except ResponseFinderError as exc:
            return ResponseUnmarshalResult(errors=[exc])

        try:
            validated_data = self._get_data(
                response.data, response.content_type, operation_response
            )
        except DataValidationError as exc:
            validated_data = None
            data_errors = [exc]
        else:
            data_errors = []

        try:
            validated_headers = self._get_headers(
                response.headers, operation_response
            )
        except HeadersError as exc:
            validated_headers = exc.headers
            headers_errors = exc.context
        else:
            headers_errors = []

        errors = list(chainiters(data_errors, headers_errors))
        return ResponseUnmarshalResult(
            errors=errors,
            data=validated_data,
            headers=validated_headers,
        )

    def _unmarshal_data(
        self,
        response: Response,
        operation: SchemaPath,
    ) -> ResponseUnmarshalResult:
        try:
            operation_response = self._find_operation_response(
                response.status_code, operation
            )
        # don't process if operation errors
        except ResponseFinderError as exc:
            return ResponseUnmarshalResult(errors=[exc])

        try:
            validated = self._get_data(
                response.data, response.content_type, operation_response
            )
        except DataValidationError as exc:
            validated = None
            data_errors = [exc]
        else:
            data_errors = []

        return ResponseUnmarshalResult(
            errors=data_errors,
            data=validated,
        )

    def _unmarshal_headers(
        self,
        response: Response,
        operation: SchemaPath,
    ) -> ResponseUnmarshalResult:
        try:
            operation_response = self._find_operation_response(
                response.status_code, operation
            )
        # don't process if operation errors
        except ResponseFinderError as exc:
            return ResponseUnmarshalResult(errors=[exc])

        try:
            validated = self._get_headers(response.headers, operation_response)
        except HeadersError as exc:
            validated = exc.headers
            headers_errors = exc.context
        else:
            headers_errors = []

        return ResponseUnmarshalResult(
            errors=headers_errors,
            headers=validated,
        )


class APICallResponseUnmarshaller(
    APICallResponseValidator, BaseResponseUnmarshaller
):
    def unmarshal(
        self,
        request: Request,
        response: Response,
    ) -> ResponseUnmarshalResult:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return ResponseUnmarshalResult(errors=[exc])

        return self._unmarshal(response, operation)


class APICallResponseDataUnmarshaller(
    APICallResponseValidator, BaseResponseUnmarshaller
):
    def unmarshal(
        self,
        request: Request,
        response: Response,
    ) -> ResponseUnmarshalResult:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return ResponseUnmarshalResult(errors=[exc])

        return self._unmarshal_data(response, operation)


class APICallResponseHeadersUnmarshaller(
    APICallResponseValidator, BaseResponseUnmarshaller
):
    def unmarshal(
        self,
        request: Request,
        response: Response,
    ) -> ResponseUnmarshalResult:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return ResponseUnmarshalResult(errors=[exc])

        return self._unmarshal_headers(response, operation)


class WebhookResponseUnmarshaller(
    WebhookResponseValidator, BaseResponseUnmarshaller
):
    def unmarshal(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> ResponseUnmarshalResult:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return ResponseUnmarshalResult(errors=[exc])

        return self._unmarshal(response, operation)


class WebhookResponseDataUnmarshaller(
    WebhookResponseValidator, BaseResponseUnmarshaller
):
    def unmarshal(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> ResponseUnmarshalResult:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return ResponseUnmarshalResult(errors=[exc])

        return self._unmarshal_data(response, operation)


class WebhookResponseHeadersUnmarshaller(
    WebhookResponseValidator, BaseResponseUnmarshaller
):
    def unmarshal(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> ResponseUnmarshalResult:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return ResponseUnmarshalResult(errors=[exc])

        return self._unmarshal_headers(response, operation)


class V30ResponseDataUnmarshaller(
    V30ResponseDataValidator, APICallResponseDataUnmarshaller
):
    schema_unmarshallers_factory = oas30_read_schema_unmarshallers_factory


class V30ResponseHeadersUnmarshaller(
    V30ResponseHeadersValidator, APICallResponseHeadersUnmarshaller
):
    schema_unmarshallers_factory = oas30_read_schema_unmarshallers_factory


class V30ResponseUnmarshaller(
    V30ResponseValidator, APICallResponseUnmarshaller
):
    schema_unmarshallers_factory = oas30_read_schema_unmarshallers_factory


class V31ResponseDataUnmarshaller(
    V31ResponseDataValidator, APICallResponseDataUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31ResponseHeadersUnmarshaller(
    V31ResponseHeadersValidator, APICallResponseHeadersUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31ResponseUnmarshaller(
    V31ResponseValidator, APICallResponseUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31WebhookResponseDataUnmarshaller(
    V31WebhookResponseDataValidator, WebhookResponseDataUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31WebhookResponseHeadersUnmarshaller(
    V31WebhookResponseHeadersValidator, WebhookResponseHeadersUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31WebhookResponseUnmarshaller(
    V31WebhookResponseValidator, WebhookResponseUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory



from typing import Optional

from jsonschema_path import SchemaPath

from openapi_core.deserializing.styles.datatypes import StyleDeserializersDict
from openapi_core.deserializing.styles.deserializers import StyleDeserializer


class StyleDeserializersFactory:
    def __init__(
        self,
        style_deserializers: Optional[StyleDeserializersDict] = None,
    ):
        if style_deserializers is None:
            style_deserializers = {}
        self.style_deserializers = style_deserializers

    def create(
        self,
        style: str,
        explode: bool,
        schema: SchemaPath,
        name: str,
    ) -> StyleDeserializer:
        schema_type = schema.getkey("type", "")

        deserialize_callable = self.style_deserializers.get(style)
        return StyleDeserializer(
            style, explode, name, schema_type, deserialize_callable
        )



from dataclasses import dataclass
from typing import Dict
from typing import Optional


@dataclass
class TemplateResult:
    pattern: str
    variables: Optional[Dict[str, str]] = None

    @property
    def resolved(self) -> str:
        if not self.variables:
            return self.pattern
        return self.pattern.format(**self.variables)



"""OpenAPI core unmarshalling request module"""

from typing import Mapping

from openapi_spec_validator.versions import consts as versions
from openapi_spec_validator.versions.datatypes import SpecVersion

from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.request.types import (
    WebhookRequestUnmarshallerType,
)
from openapi_core.unmarshalling.request.unmarshallers import (
    V30RequestUnmarshaller,
)
from openapi_core.unmarshalling.request.unmarshallers import (
    V31RequestUnmarshaller,
)
from openapi_core.unmarshalling.request.unmarshallers import (
    V31WebhookRequestUnmarshaller,
)

__all__ = [
    "UNMARSHALLERS",
    "WEBHOOK_UNMARSHALLERS",
    "V3RequestUnmarshaller",
    "V3WebhookRequestUnmarshaller",
    "V30RequestUnmarshaller",
    "V31RequestUnmarshaller",
    "V31WebhookRequestUnmarshaller",
]

# versions mapping
UNMARSHALLERS: Mapping[SpecVersion, RequestUnmarshallerType] = {
    versions.OPENAPIV30: V30RequestUnmarshaller,
    versions.OPENAPIV31: V31RequestUnmarshaller,
}
WEBHOOK_UNMARSHALLERS: Mapping[SpecVersion, WebhookRequestUnmarshallerType] = {
    versions.OPENAPIV31: V31WebhookRequestUnmarshaller,
}

# alias to the latest v3 version
V3RequestUnmarshaller = V31RequestUnmarshaller
V3WebhookRequestUnmarshaller = V31WebhookRequestUnmarshaller



from typing import Optional
from typing import cast

from jsonschema_path import SchemaPath


def get_content_type(
    prop_schema: SchemaPath, encoding: Optional[SchemaPath]
) -> str:
    if encoding is None:
        return get_default_content_type(prop_schema, encoding=False)

    if "contentType" not in encoding:
        return get_default_content_type(prop_schema, encoding=True)

    return cast(str, encoding["contentType"])


def get_default_content_type(
    prop_schema: Optional[SchemaPath], encoding: bool = False
) -> str:
    if prop_schema is None:
        return "text/plain"

    prop_type = prop_schema.getkey("type")
    if prop_type is None:
        return "text/plain" if encoding else "application/octet-stream"

    prop_format = prop_schema.getkey("format")
    if prop_type == "string" and prop_format in ["binary", "base64"]:
        return "application/octet-stream"

    if prop_type == "object":
        return "application/json"

    if prop_type == "array":
        prop_items = prop_schema / "items"
        return get_default_content_type(prop_items, encoding=encoding)

    return "text/plain"



from jsonschema_path import SchemaPath

from openapi_core.schema.servers import get_server_url


def get_spec_url(spec: SchemaPath, index: int = 0) -> str:
    servers = spec / "servers"
    return get_server_url(servers / 0)



from email.message import Message
from email.parser import Parser
from json import loads
from typing import Any
from typing import Iterator
from typing import Mapping
from typing import Tuple
from urllib.parse import parse_qsl
from xml.etree.ElementTree import Element
from xml.etree.ElementTree import fromstring

from werkzeug.datastructures import ImmutableMultiDict


def binary_loads(value: bytes, **parameters: str) -> bytes:
    return value


def plain_loads(value: bytes, **parameters: str) -> str:
    charset = "utf-8"
    if "charset" in parameters:
        charset = parameters["charset"]
    if isinstance(value, bytes):
        try:
            return value.decode(charset)
        # fallback safe decode
        except UnicodeDecodeError:
            return value.decode("ASCII", errors="surrogateescape")
    return value


def json_loads(value: bytes, **parameters: str) -> Any:
    return loads(value)


def xml_loads(value: bytes, **parameters: str) -> Element:
    charset = "utf-8"
    if "charset" in parameters:
        charset = parameters["charset"]
    return fromstring(value.decode(charset))


def urlencoded_form_loads(
    value: bytes, **parameters: str
) -> Mapping[str, Any]:
    # only UTF-8 is conforming
    return ImmutableMultiDict(parse_qsl(value.decode("utf-8")))


def data_form_loads(value: bytes, **parameters: str) -> Mapping[str, Any]:
    charset = "ASCII"
    if "charset" in parameters:
        charset = parameters["charset"]
    decoded = value.decode(charset, errors="surrogateescape")
    boundary = ""
    if "boundary" in parameters:
        boundary = parameters["boundary"]
    parser = Parser()
    mimetype = "multipart/form-data"
    header = f'Content-Type: {mimetype}; boundary="{boundary}"'
    text = "\n\n".join([header, decoded])
    parts = parser.parsestr(text, headersonly=False)
    return ImmutableMultiDict(list(iter_payloads(parts)))


def iter_payloads(parts: Message) -> Iterator[Tuple[str, bytes]]:
    for part in parts.get_payload():
        assert isinstance(part, Message)
        name = part.get_param("name", header="content-disposition")
        assert isinstance(name, str)
        payload = part.get_payload(decode=True)
        assert isinstance(payload, bytes)
        yield name, payload



"""OpenAPI core contrib falcon middlewares module"""

from typing import Any
from typing import Type
from typing import Union

from falcon.request import Request
from falcon.response import Response
from jsonschema._utils import Unset
from jsonschema.validators import _UNSET
from jsonschema_path import SchemaPath

from openapi_core import Config
from openapi_core import OpenAPI
from openapi_core.contrib.falcon.handlers import FalconOpenAPIErrorsHandler
from openapi_core.contrib.falcon.handlers import (
    FalconOpenAPIValidRequestHandler,
)
from openapi_core.contrib.falcon.integrations import FalconIntegration
from openapi_core.contrib.falcon.requests import FalconOpenAPIRequest
from openapi_core.contrib.falcon.responses import FalconOpenAPIResponse
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType


class FalconOpenAPIMiddleware(FalconIntegration):
    valid_request_handler_cls = FalconOpenAPIValidRequestHandler
    errors_handler_cls: Type[FalconOpenAPIErrorsHandler] = (
        FalconOpenAPIErrorsHandler
    )

    def __init__(
        self,
        openapi: OpenAPI,
        request_cls: Type[FalconOpenAPIRequest] = FalconOpenAPIRequest,
        response_cls: Type[FalconOpenAPIResponse] = FalconOpenAPIResponse,
        errors_handler_cls: Type[
            FalconOpenAPIErrorsHandler
        ] = FalconOpenAPIErrorsHandler,
        **unmarshaller_kwargs: Any,
    ):
        super().__init__(openapi)
        self.request_cls = request_cls or self.request_cls
        self.response_cls = response_cls or self.response_cls
        self.errors_handler_cls = errors_handler_cls or self.errors_handler_cls

    @classmethod
    def from_spec(
        cls,
        spec: SchemaPath,
        request_unmarshaller_cls: Union[
            RequestUnmarshallerType, Unset
        ] = _UNSET,
        response_unmarshaller_cls: Union[
            ResponseUnmarshallerType, Unset
        ] = _UNSET,
        request_cls: Type[FalconOpenAPIRequest] = FalconOpenAPIRequest,
        response_cls: Type[FalconOpenAPIResponse] = FalconOpenAPIResponse,
        errors_handler_cls: Type[
            FalconOpenAPIErrorsHandler
        ] = FalconOpenAPIErrorsHandler,
        **unmarshaller_kwargs: Any,
    ) -> "FalconOpenAPIMiddleware":
        config = Config(
            request_unmarshaller_cls=request_unmarshaller_cls,
            response_unmarshaller_cls=response_unmarshaller_cls,
        )
        openapi = OpenAPI(spec, config=config)
        return cls(
            openapi,
            request_unmarshaller_cls=request_unmarshaller_cls,
            response_unmarshaller_cls=response_unmarshaller_cls,
            request_cls=request_cls,
            response_cls=response_cls,
            errors_handler_cls=errors_handler_cls,
            **unmarshaller_kwargs,
        )

    def process_request(self, req: Request, resp: Response) -> None:
        valid_handler = self.valid_request_handler_cls(req, resp)
        errors_handler = self.errors_handler_cls(req, resp)
        self.handle_request(req, valid_handler, errors_handler)

    def process_response(
        self, req: Request, resp: Response, resource: Any, req_succeeded: bool
    ) -> None:
        errors_handler = self.errors_handler_cls(req, resp)
        self.handle_response(req, resp, errors_handler)



from openapi_core.deserializing.styles.datatypes import StyleDeserializersDict
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.deserializing.styles.util import deep_object_loads
from openapi_core.deserializing.styles.util import form_loads
from openapi_core.deserializing.styles.util import label_loads
from openapi_core.deserializing.styles.util import matrix_loads
from openapi_core.deserializing.styles.util import pipe_delimited_loads
from openapi_core.deserializing.styles.util import simple_loads
from openapi_core.deserializing.styles.util import space_delimited_loads

__all__ = ["style_deserializers_factory"]

style_deserializers: StyleDeserializersDict = {
    "matrix": matrix_loads,
    "label": label_loads,
    "form": form_loads,
    "simple": simple_loads,
    "spaceDelimited": space_delimited_loads,
    "pipeDelimited": pipe_delimited_loads,
    "deepObject": deep_object_loads,
}

style_deserializers_factory = StyleDeserializersFactory(
    style_deserializers=style_deserializers,
)



"""OpenAPI core contrib django module"""

from openapi_core.contrib.django.requests import DjangoOpenAPIRequest
from openapi_core.contrib.django.responses import DjangoOpenAPIResponse

__all__ = [
    "DjangoOpenAPIRequest",
    "DjangoOpenAPIResponse",
]



from dataclasses import dataclass

from openapi_core.deserializing.exceptions import DeserializeError


@dataclass
class MediaTypeDeserializeError(DeserializeError):
    """Media type deserialize operation error"""

    mimetype: str
    value: bytes

    def __str__(self) -> str:
        return (
            "Failed to deserialize value with {mimetype} mimetype: {value}"
        ).format(value=self.value.decode("utf-8"), mimetype=self.mimetype)



"""OpenAPI core unmarshalling response datatypes module"""

from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Dict
from typing import Optional

from openapi_core.unmarshalling.datatypes import BaseUnmarshalResult


@dataclass
class ResponseUnmarshalResult(BaseUnmarshalResult):
    data: Optional[str] = None
    headers: Dict[str, Any] = field(default_factory=dict)



"""OpenAPI core contrib werkzeug responses module"""

from itertools import tee

from werkzeug.datastructures import Headers
from werkzeug.wrappers import Response


class WerkzeugOpenAPIResponse:
    def __init__(self, response: Response):
        if not isinstance(response, Response):
            raise TypeError(f"'response' argument is not type of {Response}")
        self.response = response

    @property
    def data(self) -> bytes:
        if not self.response.is_sequence:
            resp_iter1, resp_iter2 = tee(self.response.iter_encoded())
            self.response.response = resp_iter1
            return b"".join(resp_iter2)
        return self.response.get_data(as_text=False)

    @property
    def status_code(self) -> int:
        return self.response._status_code

    @property
    def content_type(self) -> str:
        return str(self.response.mimetype)

    @property
    def headers(self) -> Headers:
        return Headers(self.response.headers)



"""OpenAPI core validation response validators module"""

import warnings
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional

from jsonschema_path import SchemaPath
from openapi_spec_validator import OpenAPIV30SpecValidator
from openapi_spec_validator import OpenAPIV31SpecValidator

from openapi_core.casting.schemas import oas30_read_schema_casters_factory
from openapi_core.casting.schemas import oas31_schema_casters_factory
from openapi_core.exceptions import OpenAPIError
from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.protocols import WebhookRequest
from openapi_core.templating.paths.exceptions import PathError
from openapi_core.templating.responses.exceptions import ResponseFinderError
from openapi_core.validation.decorators import ValidationErrorWrapper
from openapi_core.validation.exceptions import ValidationError
from openapi_core.validation.response.exceptions import DataValidationError
from openapi_core.validation.response.exceptions import HeadersError
from openapi_core.validation.response.exceptions import HeaderValidationError
from openapi_core.validation.response.exceptions import InvalidData
from openapi_core.validation.response.exceptions import InvalidHeader
from openapi_core.validation.response.exceptions import MissingData
from openapi_core.validation.response.exceptions import MissingHeader
from openapi_core.validation.response.exceptions import MissingRequiredHeader
from openapi_core.validation.schemas import (
    oas30_read_schema_validators_factory,
)
from openapi_core.validation.schemas import oas31_schema_validators_factory
from openapi_core.validation.validators import BaseAPICallValidator
from openapi_core.validation.validators import BaseValidator
from openapi_core.validation.validators import BaseWebhookValidator


class BaseResponseValidator(BaseValidator):
    def _iter_errors(
        self,
        status_code: int,
        data: Optional[bytes],
        headers: Mapping[str, Any],
        mimetype: str,
        operation: SchemaPath,
    ) -> Iterator[Exception]:
        try:
            operation_response = self._find_operation_response(
                status_code, operation
            )
        # don't process if operation errors
        except ResponseFinderError as exc:
            yield exc
            return

        try:
            self._get_data(data, mimetype, operation_response)
        except DataValidationError as exc:
            yield exc

        try:
            self._get_headers(headers, operation_response)
        except HeadersError as exc:
            yield from exc.context

    def _iter_data_errors(
        self,
        status_code: int,
        data: Optional[bytes],
        mimetype: str,
        operation: SchemaPath,
    ) -> Iterator[Exception]:
        try:
            operation_response = self._find_operation_response(
                status_code, operation
            )
        # don't process if operation errors
        except ResponseFinderError as exc:
            yield exc
            return

        try:
            self._get_data(data, mimetype, operation_response)
        except DataValidationError as exc:
            yield exc

    def _iter_headers_errors(
        self,
        status_code: int,
        headers: Mapping[str, Any],
        operation: SchemaPath,
    ) -> Iterator[Exception]:
        try:
            operation_response = self._find_operation_response(
                status_code, operation
            )
        # don't process if operation errors
        except ResponseFinderError as exc:
            yield exc
            return

        try:
            self._get_headers(headers, operation_response)
        except HeadersError as exc:
            yield from exc.context

    def _find_operation_response(
        self,
        status_code: int,
        operation: SchemaPath,
    ) -> SchemaPath:
        from openapi_core.templating.responses.finders import ResponseFinder

        finder = ResponseFinder(operation / "responses")
        return finder.find(str(status_code))

    @ValidationErrorWrapper(DataValidationError, InvalidData)
    def _get_data(
        self,
        data: Optional[bytes],
        mimetype: str,
        operation_response: SchemaPath,
    ) -> Any:
        if "content" not in operation_response:
            return None

        content = operation_response / "content"

        raw_data = self._get_data_value(data)
        value, _ = self._get_content_and_schema(raw_data, content, mimetype)
        return value

    def _get_data_value(self, data: Optional[bytes]) -> bytes:
        if not data:
            raise MissingData

        return data

    def _get_headers(
        self, headers: Mapping[str, Any], operation_response: SchemaPath
    ) -> Dict[str, Any]:
        if "headers" not in operation_response:
            return {}

        response_headers = operation_response / "headers"

        errors: List[OpenAPIError] = []
        validated: Dict[str, Any] = {}
        for name, header in list(response_headers.items()):
            # ignore Content-Type header
            if name.lower() == "content-type":
                continue
            try:
                value = self._get_header(headers, name, header)
            except MissingHeader:
                continue
            except ValidationError as exc:
                errors.append(exc)
                continue
            else:
                validated[name] = value

        if errors:
            raise HeadersError(context=iter(errors), headers=validated)

        return validated

    @ValidationErrorWrapper(HeaderValidationError, InvalidHeader, name="name")
    def _get_header(
        self, headers: Mapping[str, Any], name: str, header: SchemaPath
    ) -> Any:
        deprecated = header.getkey("deprecated", False)
        if deprecated:
            warnings.warn(
                f"{name} header is deprecated",
                DeprecationWarning,
            )

        try:
            value, _ = self._get_param_or_header_and_schema(
                header, headers, name=name
            )
        except KeyError:
            required = header.getkey("required", False)
            if required:
                raise MissingRequiredHeader(name)
            raise MissingHeader(name)
        else:
            return value


class BaseAPICallResponseValidator(
    BaseResponseValidator, BaseAPICallValidator
):
    def iter_errors(
        self,
        request: Request,
        response: Response,
    ) -> Iterator[Exception]:
        raise NotImplementedError

    def validate(
        self,
        request: Request,
        response: Response,
    ) -> None:
        for err in self.iter_errors(request, response):
            raise err


class BaseWebhookResponseValidator(
    BaseResponseValidator, BaseWebhookValidator
):
    def iter_errors(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> Iterator[Exception]:
        raise NotImplementedError

    def validate(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> None:
        for err in self.iter_errors(request, response):
            raise err


class APICallResponseDataValidator(BaseAPICallResponseValidator):
    def iter_errors(
        self,
        request: Request,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_data_errors(
            response.status_code,
            response.data,
            response.content_type,
            operation,
        )


class APICallResponseHeadersValidator(BaseAPICallResponseValidator):
    def iter_errors(
        self,
        request: Request,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_headers_errors(
            response.status_code, response.headers, operation
        )


class APICallResponseValidator(BaseAPICallResponseValidator):
    def iter_errors(
        self,
        request: Request,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_errors(
            response.status_code,
            response.data,
            response.headers,
            response.content_type,
            operation,
        )


class WebhookResponseDataValidator(BaseWebhookResponseValidator):
    def iter_errors(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_data_errors(
            response.status_code,
            response.data,
            response.content_type,
            operation,
        )


class WebhookResponseHeadersValidator(BaseWebhookResponseValidator):
    def iter_errors(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_headers_errors(
            response.status_code, response.headers, operation
        )


class WebhookResponseValidator(BaseWebhookResponseValidator):
    def iter_errors(
        self,
        request: WebhookRequest,
        response: Response,
    ) -> Iterator[Exception]:
        try:
            _, operation, _, _, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            yield exc
            return

        yield from self._iter_errors(
            response.status_code,
            response.data,
            response.headers,
            response.content_type,
            operation,
        )


class V30ResponseDataValidator(APICallResponseDataValidator):
    spec_validator_cls = OpenAPIV30SpecValidator
    schema_casters_factory = oas30_read_schema_casters_factory
    schema_validators_factory = oas30_read_schema_validators_factory


class V30ResponseHeadersValidator(APICallResponseHeadersValidator):
    spec_validator_cls = OpenAPIV30SpecValidator
    schema_casters_factory = oas30_read_schema_casters_factory
    schema_validators_factory = oas30_read_schema_validators_factory


class V30ResponseValidator(APICallResponseValidator):
    spec_validator_cls = OpenAPIV30SpecValidator
    schema_casters_factory = oas30_read_schema_casters_factory
    schema_validators_factory = oas30_read_schema_validators_factory


class V31ResponseDataValidator(APICallResponseDataValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31ResponseHeadersValidator(APICallResponseHeadersValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31ResponseValidator(APICallResponseValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookResponseDataValidator(WebhookResponseDataValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookResponseHeadersValidator(WebhookResponseHeadersValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory


class V31WebhookResponseValidator(WebhookResponseValidator):
    spec_validator_cls = OpenAPIV31SpecValidator
    schema_casters_factory = oas31_schema_casters_factory
    schema_validators_factory = oas31_schema_validators_factory



from typing import Any
from typing import Optional

from parse import Match
from parse import Parser


class ExtendedParser(Parser):  # type: ignore
    def _handle_field(self, field: str) -> Any:
        # handle as path parameter field
        field = field[1:-1]
        path_parameter_field = "{%s:PathParameter}" % field
        return super()._handle_field(path_parameter_field)


class PathParameter:
    name = "PathParameter"
    pattern = r"[^\/]*"

    def __call__(self, text: str) -> str:
        return text


parse_path_parameter = PathParameter()


def search(path_pattern: str, full_url_pattern: str) -> Optional[Match]:
    extra_types = {parse_path_parameter.name: parse_path_parameter}
    p = ExtendedParser(path_pattern, extra_types)
    p._expression = p._expression + "$"
    return p.search(full_url_pattern)


def parse(server_url: str, server_url_pattern: str) -> Match:
    extra_types = {parse_path_parameter.name: parse_path_parameter}
    p = ExtendedParser(server_url, extra_types)
    p._expression = "^" + p._expression
    return p.parse(server_url_pattern)



"""OpenAPI core contrib aiohttp requests module"""

from __future__ import annotations

from aiohttp import web

from openapi_core.datatypes import RequestParameters


class Empty: ...


_empty = Empty()


class AIOHTTPOpenAPIWebRequest:
    __slots__ = ("request", "parameters", "_get_body", "_body")

    def __init__(self, request: web.Request, *, body: bytes | None):
        if not isinstance(request, web.Request):
            raise TypeError(
                f"'request' argument is not type of {web.Request.__qualname__!r}"
            )
        self.request = request
        self.parameters = RequestParameters(
            query=self.request.query,
            header=self.request.headers,
            cookie=self.request.cookies,
        )
        self._body = body

    @property
    def host_url(self) -> str:
        return f"{self.request.url.scheme}://{self.request.url.host}"

    @property
    def path(self) -> str:
        return self.request.url.path

    @property
    def method(self) -> str:
        return self.request.method.lower()

    @property
    def body(self) -> bytes | None:
        return self._body

    @property
    def content_type(self) -> str:
        return self.request.content_type



from typing import Optional

from openapi_core.datatypes import Parameters


class ResultMock:
    def __init__(
        self,
        body: Optional[str] = None,
        parameters: Optional[Parameters] = None,
        data: Optional[str] = None,
        error_to_raise: Optional[Exception] = None,
    ):
        self.body = body
        self.parameters = parameters
        self.data = data
        self.error_to_raise = error_to_raise

    def raise_for_errors(self) -> None:
        if self.error_to_raise is not None:
            raise self.error_to_raise



"""OpenAPI core validation exceptions module"""

from dataclasses import dataclass

from openapi_core.exceptions import OpenAPIError


@dataclass
class ValidationError(OpenAPIError):
    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.__cause__}"



"""OpenAPI core contrib starlette responses module"""

from typing import Optional

from starlette.datastructures import Headers
from starlette.responses import Response
from starlette.responses import StreamingResponse


class StarletteOpenAPIResponse:
    def __init__(self, response: Response, data: Optional[bytes] = None):
        if not isinstance(response, Response):
            raise TypeError(f"'response' argument is not type of {Response}")
        self.response = response

        if data is None and isinstance(response, StreamingResponse):
            raise RuntimeError(
                f"'data' argument is required for {StreamingResponse}"
            )
        self._data = data

    @property
    def data(self) -> bytes:
        if self._data is not None:
            return self._data
        if isinstance(self.response.body, bytes):
            return self.response.body
        assert isinstance(self.response.body, str)
        return self.response.body.encode("utf-8")

    @property
    def status_code(self) -> int:
        return self.response.status_code

    @property
    def content_type(self) -> str:
        return self.response.headers.get("Content-Type") or ""

    @property
    def headers(self) -> Headers:
        return self.response.headers



import warnings
from typing import Any

from jsonschema_path import SchemaPath

from openapi_core.datatypes import RequestParameters
from openapi_core.security.exceptions import SecurityProviderError


class BaseProvider:
    def __init__(self, scheme: SchemaPath):
        self.scheme = scheme

    def __call__(self, parameters: RequestParameters) -> Any:
        raise NotImplementedError


class UnsupportedProvider(BaseProvider):
    def __call__(self, parameters: RequestParameters) -> Any:
        warnings.warn("Unsupported scheme type")


class ApiKeyProvider(BaseProvider):
    def __call__(self, parameters: RequestParameters) -> Any:
        name = self.scheme["name"]
        location = self.scheme["in"]
        source = getattr(parameters, location)
        if name not in source:
            raise SecurityProviderError("Missing api key parameter.")
        return source[name]


class HttpProvider(BaseProvider):
    def __call__(self, parameters: RequestParameters) -> Any:
        if "Authorization" not in parameters.header:
            raise SecurityProviderError("Missing authorization header.")
        auth_header = parameters.header["Authorization"]
        try:
            auth_type, encoded_credentials = auth_header.split(" ", 1)
        except ValueError:
            raise SecurityProviderError(
                "Could not parse authorization header."
            )

        scheme = self.scheme["scheme"]
        if auth_type.lower() != scheme:
            raise SecurityProviderError(
                f"Unknown authorization method {auth_type}"
            )

        return encoded_credentials



"""OpenAPI core schemas util module"""

from base64 import b64decode
from datetime import date
from datetime import datetime
from typing import Any
from typing import Union
from uuid import UUID


def format_date(value: str) -> date:
    return datetime.strptime(value, "%Y-%m-%d").date()


def format_uuid(value: Any) -> UUID:
    if isinstance(value, UUID):
        return value
    return UUID(value)


def format_byte(value: str, encoding: str = "utf8") -> str:
    return str(b64decode(value), encoding)


def format_number(value: str) -> Union[int, float]:
    if isinstance(value, (int, float)):
        return value

    return float(value)



from typing import Any
from typing import Optional

from jsonschema_path import SchemaPath

from openapi_core.protocols import Request
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.request.protocols import RequestUnmarshaller
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType


class RequestUnmarshallingProcessor:
    def __init__(
        self,
        spec: SchemaPath,
        request_unmarshaller_cls: RequestUnmarshallerType,
        **unmarshaller_kwargs: Any
    ) -> None:
        self.spec = spec
        self.request_unmarshaller_cls = request_unmarshaller_cls
        self.unmarshaller_kwargs = unmarshaller_kwargs

        self._request_unmarshaller_cached: Optional[RequestUnmarshaller] = None

    @property
    def request_unmarshaller(self) -> RequestUnmarshaller:
        if self._request_unmarshaller_cached is None:
            self._request_unmarshaller_cached = self.request_unmarshaller_cls(
                self.spec, **self.unmarshaller_kwargs
            )
        return self._request_unmarshaller_cached

    def process(self, request: Request) -> RequestUnmarshalResult:
        return self.request_unmarshaller.unmarshal(request)



"""OpenAPI X-Model extension factories module"""

from dataclasses import make_dataclass
from pydoc import locate
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Type

from jsonschema_path import SchemaPath

from openapi_core.extensions.models.types import Field


class DictFactory:
    base_class = dict

    def create(
        self, schema: SchemaPath, fields: Iterable[Field]
    ) -> Type[Dict[Any, Any]]:
        return self.base_class


class ModelFactory(DictFactory):
    def create(
        self,
        schema: SchemaPath,
        fields: Iterable[Field],
    ) -> Type[Any]:
        name = schema.getkey("x-model")
        if name is None:
            return super().create(schema, fields)

        return make_dataclass(name, fields, frozen=True)


class ModelPathFactory(ModelFactory):
    def create(
        self,
        schema: SchemaPath,
        fields: Iterable[Field],
    ) -> Any:
        model_class_path = schema.getkey("x-model-path")
        if model_class_path is None:
            return super().create(schema, fields)

        return locate(model_class_path)



"""OpenAPI core contrib flask requests module"""

from flask.wrappers import Request
from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict

from openapi_core.contrib.werkzeug.requests import WerkzeugOpenAPIRequest
from openapi_core.datatypes import RequestParameters


class FlaskOpenAPIRequest(WerkzeugOpenAPIRequest):
    def __init__(self, request: Request):
        if not isinstance(request, Request):
            raise TypeError(f"'request' argument is not type of {Request}")
        self.request: Request = request

        self.parameters = RequestParameters(
            path=self.request.view_args or {},
            query=ImmutableMultiDict(self.request.args),
            header=Headers(self.request.headers),
            cookie=self.request.cookies,
        )

    @property
    def path_pattern(self) -> str:
        if self.request.url_rule is None:
            return self.path

        path = self.get_path(self.request.url_rule.rule)
        return self.path_regex.sub(r"{\1}", path)



from typing import Mapping
from typing import Optional

from jsonschema_path import SchemaPath

from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.deserializers import (
    MediaTypeDeserializer,
)
from openapi_core.deserializing.media_types.deserializers import (
    MediaTypesDeserializer,
)
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)


class MediaTypeDeserializersFactory:
    def __init__(
        self,
        style_deserializers_factory: StyleDeserializersFactory,
        media_type_deserializers: Optional[MediaTypeDeserializersDict] = None,
    ):
        self.style_deserializers_factory = style_deserializers_factory
        if media_type_deserializers is None:
            media_type_deserializers = {}
        self.media_type_deserializers = media_type_deserializers

    def create(
        self,
        mimetype: str,
        schema: Optional[SchemaPath] = None,
        parameters: Optional[Mapping[str, str]] = None,
        encoding: Optional[SchemaPath] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
    ) -> MediaTypeDeserializer:
        if parameters is None:
            parameters = {}
        if extra_media_type_deserializers is None:
            extra_media_type_deserializers = {}
        media_types_deserializer = MediaTypesDeserializer(
            self.media_type_deserializers,
            extra_media_type_deserializers,
        )

        return MediaTypeDeserializer(
            self.style_deserializers_factory,
            media_types_deserializer,
            mimetype,
            schema=schema,
            encoding=encoding,
            **parameters,
        )



"""OpenAPI core unmarshalling response module"""

from typing import Mapping

from openapi_spec_validator.versions import consts as versions
from openapi_spec_validator.versions.datatypes import SpecVersion

from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType
from openapi_core.unmarshalling.response.types import (
    WebhookResponseUnmarshallerType,
)
from openapi_core.unmarshalling.response.unmarshallers import (
    V30ResponseUnmarshaller,
)
from openapi_core.unmarshalling.response.unmarshallers import (
    V31ResponseUnmarshaller,
)
from openapi_core.unmarshalling.response.unmarshallers import (
    V31WebhookResponseUnmarshaller,
)

__all__ = [
    "UNMARSHALLERS",
    "WEBHOOK_UNMARSHALLERS",
    "V3ResponseUnmarshaller",
    "V3WebhookResponseUnmarshaller",
    "V30ResponseUnmarshaller",
    "V31ResponseUnmarshaller",
    "V31WebhookResponseUnmarshaller",
]

# versions mapping
UNMARSHALLERS: Mapping[SpecVersion, ResponseUnmarshallerType] = {
    versions.OPENAPIV30: V30ResponseUnmarshaller,
    versions.OPENAPIV31: V31ResponseUnmarshaller,
}
WEBHOOK_UNMARSHALLERS: Mapping[
    SpecVersion, WebhookResponseUnmarshallerType
] = {
    versions.OPENAPIV31: V31WebhookResponseUnmarshaller,
}

# alias to the latest v3 version
V3ResponseUnmarshaller = V31ResponseUnmarshaller
V3WebhookResponseUnmarshaller = V31WebhookResponseUnmarshaller



from falcon.request import Request
from falcon.response import Response

from openapi_core.contrib.falcon.requests import FalconOpenAPIRequest
from openapi_core.contrib.falcon.responses import FalconOpenAPIResponse
from openapi_core.unmarshalling.processors import UnmarshallingProcessor
from openapi_core.unmarshalling.typing import ErrorsHandlerCallable


class FalconIntegration(UnmarshallingProcessor[Request, Response]):
    request_cls = FalconOpenAPIRequest
    response_cls = FalconOpenAPIResponse

    def get_openapi_request(self, request: Request) -> FalconOpenAPIRequest:
        return self.request_cls(request)

    def get_openapi_response(
        self, response: Response
    ) -> FalconOpenAPIResponse:
        assert self.response_cls is not None
        return self.response_cls(response)

    def should_validate_response(self) -> bool:
        return self.response_cls is not None

    def handle_response(
        self,
        request: Request,
        response: Response,
        errors_handler: ErrorsHandlerCallable[Response],
    ) -> Response:
        if not self.should_validate_response():
            return response
        return super().handle_response(request, response, errors_handler)



from typing import Type
from typing import Union

from openapi_core.validation.response.protocols import ResponseValidator
from openapi_core.validation.response.protocols import WebhookResponseValidator

ResponseValidatorType = Type[ResponseValidator]
WebhookResponseValidatorType = Type[WebhookResponseValidator]
AnyResponseValidatorType = Union[
    ResponseValidatorType, WebhookResponseValidatorType
]



from openapi_core.contrib.flask.decorators import FlaskOpenAPIViewDecorator
from openapi_core.contrib.flask.requests import FlaskOpenAPIRequest
from openapi_core.contrib.flask.responses import FlaskOpenAPIResponse

__all__ = [
    "FlaskOpenAPIViewDecorator",
    "FlaskOpenAPIRequest",
    "FlaskOpenAPIResponse",
]



from dataclasses import dataclass
from typing import Union

from jsonschema._utils import Unset
from jsonschema.validators import _UNSET
from openapi_spec_validator.validation.types import SpecValidatorType

from openapi_core.unmarshalling.configurations import UnmarshallerConfig
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.request.types import (
    WebhookRequestUnmarshallerType,
)
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType
from openapi_core.unmarshalling.response.types import (
    WebhookResponseUnmarshallerType,
)
from openapi_core.validation.request.types import RequestValidatorType
from openapi_core.validation.request.types import WebhookRequestValidatorType
from openapi_core.validation.response.types import ResponseValidatorType
from openapi_core.validation.response.types import WebhookResponseValidatorType


@dataclass
class Config(UnmarshallerConfig):
    """OpenAPI configuration dataclass.

    Read more information, in the
    [OpenAPI-core docs for Configuration](https://openapi-core.readthedocs.io/configuration/).

    Attributes:
        spec_validator_cls: Specification validator class.
        spec_base_uri: Specification base URI. Deprecated, use base_uri parameter in OpenAPI.from_dict and OpenAPI.from_file if you want to define it.
        request_validator_cls: Request validator class.
        response_validator_cls: Response validator class.
        webhook_request_validator_cls: Webhook request validator class.
        webhook_response_validator_cls: Webhook response validator class.
        request_unmarshaller_cls: Request unmarshaller class.
        response_unmarshaller_cls: Response unmarshaller class.
        webhook_request_unmarshaller_cls: Webhook request unmarshaller class.
        webhook_response_unmarshaller_cls: Webhook response unmarshaller class.
    """

    spec_validator_cls: Union[SpecValidatorType, Unset] = _UNSET
    spec_base_uri: str = ""

    request_validator_cls: Union[RequestValidatorType, Unset] = _UNSET
    response_validator_cls: Union[ResponseValidatorType, Unset] = _UNSET
    webhook_request_validator_cls: Union[
        WebhookRequestValidatorType, Unset
    ] = _UNSET
    webhook_response_validator_cls: Union[
        WebhookResponseValidatorType, Unset
    ] = _UNSET
    request_unmarshaller_cls: Union[RequestUnmarshallerType, Unset] = _UNSET
    response_unmarshaller_cls: Union[ResponseUnmarshallerType, Unset] = _UNSET
    webhook_request_unmarshaller_cls: Union[
        WebhookRequestUnmarshallerType, Unset
    ] = _UNSET
    webhook_response_unmarshaller_cls: Union[
        WebhookResponseUnmarshallerType, Unset
    ] = _UNSET



"""OpenAPI core contrib django decorators module"""

from typing import Any
from typing import Callable
from typing import Optional
from typing import Type

from django.conf import settings
from django.http.request import HttpRequest
from django.http.response import HttpResponse
from jsonschema_path import SchemaPath

from openapi_core import OpenAPI
from openapi_core.contrib.django.handlers import DjangoOpenAPIErrorsHandler
from openapi_core.contrib.django.handlers import (
    DjangoOpenAPIValidRequestHandler,
)
from openapi_core.contrib.django.integrations import DjangoIntegration
from openapi_core.contrib.django.providers import get_default_openapi_instance
from openapi_core.contrib.django.requests import DjangoOpenAPIRequest
from openapi_core.contrib.django.responses import DjangoOpenAPIResponse


class DjangoOpenAPIViewDecorator(DjangoIntegration):
    valid_request_handler_cls = DjangoOpenAPIValidRequestHandler
    errors_handler_cls: Type[DjangoOpenAPIErrorsHandler] = (
        DjangoOpenAPIErrorsHandler
    )

    def __init__(
        self,
        openapi: Optional[OpenAPI] = None,
        request_cls: Type[DjangoOpenAPIRequest] = DjangoOpenAPIRequest,
        response_cls: Type[DjangoOpenAPIResponse] = DjangoOpenAPIResponse,
        errors_handler_cls: Type[
            DjangoOpenAPIErrorsHandler
        ] = DjangoOpenAPIErrorsHandler,
    ):
        if openapi is None:
            openapi = get_default_openapi_instance()

        super().__init__(openapi)

        # If OPENAPI_RESPONSE_CLS is defined in settings.py (for custom response classes),
        # set the response_cls accordingly.
        if hasattr(settings, "OPENAPI_RESPONSE_CLS"):
            response_cls = settings.OPENAPI_RESPONSE_CLS

        self.request_cls = request_cls
        self.response_cls = response_cls

    def __call__(self, view_func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Thanks to this method, the class acts as a decorator.
        Example usage:

            @DjangoOpenAPIViewDecorator()
            def my_view(request): ...

        """

        def _wrapped_view(
            request: HttpRequest, *args: Any, **kwargs: Any
        ) -> HttpResponse:
            # get_response is the function that we treats
            # as the "next step" in the chain (i.e., our original view).
            def get_response(r: HttpRequest) -> HttpResponse:
                return view_func(r, *args, **kwargs)

            # Create a handler that will validate the request.
            valid_request_handler = self.valid_request_handler_cls(
                request, get_response
            )

            # Validate the request (before running the view).
            errors_handler = self.errors_handler_cls()
            response = self.handle_request(
                request, valid_request_handler, errors_handler
            )

            # Validate the response (after the view) if should_validate_response() returns True.
            return self.handle_response(request, response, errors_handler)

        return _wrapped_view

    @classmethod
    def from_spec(
        cls,
        spec: SchemaPath,
        request_cls: Type[DjangoOpenAPIRequest] = DjangoOpenAPIRequest,
        response_cls: Type[DjangoOpenAPIResponse] = DjangoOpenAPIResponse,
        errors_handler_cls: Type[
            DjangoOpenAPIErrorsHandler
        ] = DjangoOpenAPIErrorsHandler,
    ) -> "DjangoOpenAPIViewDecorator":
        openapi = OpenAPI(spec)
        return cls(
            openapi,
            request_cls=request_cls,
            response_cls=response_cls,
            errors_handler_cls=errors_handler_cls,
        )



from collections import OrderedDict

from isodate.isodatetime import parse_datetime

from openapi_core.unmarshalling.schemas.factories import (
    SchemaUnmarshallersFactory,
)
from openapi_core.unmarshalling.schemas.unmarshallers import AnyUnmarshaller
from openapi_core.unmarshalling.schemas.unmarshallers import ArrayUnmarshaller
from openapi_core.unmarshalling.schemas.unmarshallers import (
    MultiTypeUnmarshaller,
)
from openapi_core.unmarshalling.schemas.unmarshallers import ObjectUnmarshaller
from openapi_core.unmarshalling.schemas.unmarshallers import (
    PrimitiveUnmarshaller,
)
from openapi_core.unmarshalling.schemas.unmarshallers import TypesUnmarshaller
from openapi_core.unmarshalling.schemas.util import format_byte
from openapi_core.unmarshalling.schemas.util import format_date
from openapi_core.unmarshalling.schemas.util import format_uuid
from openapi_core.validation.schemas import (
    oas30_read_schema_validators_factory,
)
from openapi_core.validation.schemas import (
    oas30_write_schema_validators_factory,
)
from openapi_core.validation.schemas import oas31_schema_validators_factory

__all__ = [
    "oas30_format_unmarshallers",
    "oas31_format_unmarshallers",
    "oas30_write_schema_unmarshallers_factory",
    "oas30_read_schema_unmarshallers_factory",
    "oas31_schema_unmarshallers_factory",
]

oas30_unmarshallers_dict = OrderedDict(
    [
        ("object", ObjectUnmarshaller),
        ("array", ArrayUnmarshaller),
        ("boolean", PrimitiveUnmarshaller),
        ("integer", PrimitiveUnmarshaller),
        ("number", PrimitiveUnmarshaller),
        ("string", PrimitiveUnmarshaller),
    ]
)
oas31_unmarshallers_dict = oas30_unmarshallers_dict.copy()
oas31_unmarshallers_dict.update(
    {
        "null": PrimitiveUnmarshaller,
    }
)

oas30_types_unmarshaller = TypesUnmarshaller(
    oas30_unmarshallers_dict,
    AnyUnmarshaller,
)
oas31_types_unmarshaller = TypesUnmarshaller(
    oas31_unmarshallers_dict,
    AnyUnmarshaller,
    multi=MultiTypeUnmarshaller,
)

oas30_format_unmarshallers = {
    # string compatible
    "date": format_date,
    "date-time": parse_datetime,
    "binary": bytes,
    "uuid": format_uuid,
    "byte": format_byte,
}
oas31_format_unmarshallers = oas30_format_unmarshallers

oas30_write_schema_unmarshallers_factory = SchemaUnmarshallersFactory(
    oas30_write_schema_validators_factory,
    oas30_types_unmarshaller,
    format_unmarshallers=oas30_format_unmarshallers,
)

oas30_read_schema_unmarshallers_factory = SchemaUnmarshallersFactory(
    oas30_read_schema_validators_factory,
    oas30_types_unmarshaller,
    format_unmarshallers=oas30_format_unmarshallers,
)

oas31_schema_unmarshallers_factory = SchemaUnmarshallersFactory(
    oas31_schema_validators_factory,
    oas31_types_unmarshaller,
    format_unmarshallers=oas31_format_unmarshallers,
)

# alias to v31 version (request/response are the same bcs no context needed)
oas31_request_schema_unmarshallers_factory = oas31_schema_unmarshallers_factory
oas31_response_schema_unmarshallers_factory = (
    oas31_schema_unmarshallers_factory
)



import warnings
from typing import Any
from typing import Mapping
from typing import Optional

from openapi_core.deserializing.exceptions import DeserializeError
from openapi_core.deserializing.styles.datatypes import DeserializerCallable


class StyleDeserializer:
    def __init__(
        self,
        style: str,
        explode: bool,
        name: str,
        schema_type: str,
        deserializer_callable: Optional[DeserializerCallable] = None,
    ):
        self.style = style
        self.explode = explode
        self.name = name
        self.schema_type = schema_type
        self.deserializer_callable = deserializer_callable

    def deserialize(self, location: Mapping[str, Any]) -> Any:
        if self.deserializer_callable is None:
            warnings.warn(f"Unsupported {self.style} style")
            return location[self.name]

        try:
            return self.deserializer_callable(
                self.explode, self.name, self.schema_type, location
            )
        except (ValueError, TypeError, AttributeError):
            raise DeserializeError(self.style, self.name)



"""OpenAPI core contrib starlette middlewares module"""

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from openapi_core import OpenAPI
from openapi_core.contrib.starlette.handlers import (
    StarletteOpenAPIErrorsHandler,
)
from openapi_core.contrib.starlette.handlers import (
    StarletteOpenAPIValidRequestHandler,
)
from openapi_core.contrib.starlette.integrations import StarletteIntegration


class StarletteOpenAPIMiddleware(StarletteIntegration, BaseHTTPMiddleware):
    valid_request_handler_cls = StarletteOpenAPIValidRequestHandler
    errors_handler = StarletteOpenAPIErrorsHandler()

    def __init__(self, app: ASGIApp, openapi: OpenAPI):
        super().__init__(openapi)
        BaseHTTPMiddleware.__init__(self, app)

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        valid_request_handler = self.valid_request_handler_cls(
            request, call_next
        )
        response = await self.handle_request(
            request, valid_request_handler, self.errors_handler
        )
        return await self.handle_response(
            request, response, self.errors_handler
        )



from openapi_core.contrib.falcon.requests import FalconOpenAPIRequest
from openapi_core.contrib.falcon.responses import FalconOpenAPIResponse

__all__ = [
    "FalconOpenAPIRequest",
    "FalconOpenAPIResponse",
]



from typing import Any
from typing import Mapping
from typing import Optional
from typing import Tuple

from jsonschema_path import SchemaPath
from openapi_spec_validator.validation.types import SpecValidatorType

from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.deserializing.media_types import (
    media_type_deserializers_factory,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.styles import style_deserializers_factory
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.templating.paths.types import PathFinderType
from openapi_core.unmarshalling.schemas.datatypes import (
    FormatUnmarshallersDict,
)
from openapi_core.unmarshalling.schemas.factories import (
    SchemaUnmarshallersFactory,
)
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory
from openapi_core.validation.validators import BaseValidator


class BaseUnmarshaller(BaseValidator):
    schema_unmarshallers_factory: SchemaUnmarshallersFactory = NotImplemented

    def __init__(
        self,
        spec: SchemaPath,
        base_url: Optional[str] = None,
        style_deserializers_factory: StyleDeserializersFactory = style_deserializers_factory,
        media_type_deserializers_factory: MediaTypeDeserializersFactory = media_type_deserializers_factory,
        schema_casters_factory: Optional[SchemaCastersFactory] = None,
        schema_validators_factory: Optional[SchemaValidatorsFactory] = None,
        path_finder_cls: Optional[PathFinderType] = None,
        spec_validator_cls: Optional[SpecValidatorType] = None,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
        schema_unmarshallers_factory: Optional[
            SchemaUnmarshallersFactory
        ] = None,
        format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
        extra_format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
    ):
        if schema_validators_factory is None and schema_unmarshallers_factory:
            schema_validators_factory = (
                schema_unmarshallers_factory.schema_validators_factory
            )
        BaseValidator.__init__(
            self,
            spec,
            base_url=base_url,
            style_deserializers_factory=style_deserializers_factory,
            media_type_deserializers_factory=media_type_deserializers_factory,
            schema_casters_factory=schema_casters_factory,
            schema_validators_factory=schema_validators_factory,
            path_finder_cls=path_finder_cls,
            spec_validator_cls=spec_validator_cls,
            format_validators=format_validators,
            extra_format_validators=extra_format_validators,
            extra_media_type_deserializers=extra_media_type_deserializers,
        )
        self.schema_unmarshallers_factory = (
            schema_unmarshallers_factory or self.schema_unmarshallers_factory
        )
        if self.schema_unmarshallers_factory is NotImplemented:
            raise NotImplementedError(
                "schema_unmarshallers_factory is not assigned"
            )
        self.format_unmarshallers = format_unmarshallers
        self.extra_format_unmarshallers = extra_format_unmarshallers

    def _unmarshal_schema(self, schema: SchemaPath, value: Any) -> Any:
        unmarshaller = self.schema_unmarshallers_factory.create(
            schema,
            format_validators=self.format_validators,
            extra_format_validators=self.extra_format_validators,
            format_unmarshallers=self.format_unmarshallers,
            extra_format_unmarshallers=self.extra_format_unmarshallers,
        )
        return unmarshaller.unmarshal(value)

    def _get_param_or_header_and_schema(
        self,
        param_or_header: SchemaPath,
        location: Mapping[str, Any],
        name: Optional[str] = None,
    ) -> Tuple[Any, Optional[SchemaPath]]:
        casted, schema = super()._get_param_or_header_and_schema(
            param_or_header, location, name=name
        )
        if schema is None:
            return casted, None
        return self._unmarshal_schema(schema, casted), schema

    def _get_content_and_schema(
        self, raw: Any, content: SchemaPath, mimetype: Optional[str] = None
    ) -> Tuple[Any, Optional[SchemaPath]]:
        casted, schema = super()._get_content_and_schema(
            raw, content, mimetype
        )
        if schema is None:
            return casted, None
        return self._unmarshal_schema(schema, casted), schema



from typing import Any
from typing import Mapping
from typing import Optional
from xml.etree.ElementTree import ParseError

from jsonschema_path import SchemaPath

from openapi_core.deserializing.media_types.datatypes import (
    DeserializerCallable,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.exceptions import (
    MediaTypeDeserializeError,
)
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.schema.encodings import get_content_type
from openapi_core.schema.parameters import get_style_and_explode
from openapi_core.schema.protocols import SuportsGetAll
from openapi_core.schema.protocols import SuportsGetList
from openapi_core.schema.schemas import get_properties


class MediaTypesDeserializer:
    def __init__(
        self,
        media_type_deserializers: Optional[MediaTypeDeserializersDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
    ):
        if media_type_deserializers is None:
            media_type_deserializers = {}
        self.media_type_deserializers = media_type_deserializers
        if extra_media_type_deserializers is None:
            extra_media_type_deserializers = {}
        self.extra_media_type_deserializers = extra_media_type_deserializers

    def deserialize(
        self, mimetype: str, value: bytes, **parameters: str
    ) -> Any:
        deserializer_callable = self.get_deserializer_callable(mimetype)

        try:
            return deserializer_callable(value, **parameters)
        except (ParseError, ValueError, TypeError, AttributeError):
            raise MediaTypeDeserializeError(mimetype, value)

    def get_deserializer_callable(
        self,
        mimetype: str,
    ) -> DeserializerCallable:
        if mimetype in self.extra_media_type_deserializers:
            return self.extra_media_type_deserializers[mimetype]
        return self.media_type_deserializers[mimetype]


class MediaTypeDeserializer:
    def __init__(
        self,
        style_deserializers_factory: StyleDeserializersFactory,
        media_types_deserializer: MediaTypesDeserializer,
        mimetype: str,
        schema: Optional[SchemaPath] = None,
        encoding: Optional[SchemaPath] = None,
        **parameters: str,
    ):
        self.style_deserializers_factory = style_deserializers_factory
        self.media_types_deserializer = media_types_deserializer
        self.mimetype = mimetype
        self.schema = schema
        self.encoding = encoding
        self.parameters = parameters

    def deserialize(self, value: bytes) -> Any:
        deserialized = self.media_types_deserializer.deserialize(
            self.mimetype, value, **self.parameters
        )

        if (
            self.mimetype != "application/x-www-form-urlencoded"
            and not self.mimetype.startswith("multipart")
        ):
            return deserialized

        # decode multipart request bodies
        return self.decode(deserialized)

    def evolve(
        self, mimetype: str, schema: Optional[SchemaPath]
    ) -> "MediaTypeDeserializer":
        cls = self.__class__

        return cls(
            self.style_deserializers_factory,
            self.media_types_deserializer,
            mimetype,
            schema=schema,
        )

    def decode(self, location: Mapping[str, Any]) -> Mapping[str, Any]:
        # schema is required for multipart
        assert self.schema is not None
        properties = {}
        for prop_name, prop_schema in get_properties(self.schema).items():
            try:
                properties[prop_name] = self.decode_property(
                    prop_name, prop_schema, location
                )
            except KeyError:
                if "default" not in prop_schema:
                    continue
                properties[prop_name] = prop_schema["default"]

        return properties

    def decode_property(
        self,
        prop_name: str,
        prop_schema: SchemaPath,
        location: Mapping[str, Any],
    ) -> Any:
        if self.encoding is None or prop_name not in self.encoding:
            if self.mimetype == "application/x-www-form-urlencoded":
                # default serialization strategy for complex objects
                # in the application/x-www-form-urlencoded
                return self.decode_property_style(
                    prop_name,
                    prop_schema,
                    location,
                    SchemaPath.from_dict({"style": "form"}),
                )
            return self.decode_property_content_type(
                prop_name, prop_schema, location
            )

        prep_encoding = self.encoding / prop_name
        if (
            "style" not in prep_encoding
            and "explode" not in prep_encoding
            and "allowReserved" not in prep_encoding
        ):
            return self.decode_property_content_type(
                prop_name, prop_schema, location, prep_encoding
            )

        return self.decode_property_style(
            prop_name, prop_schema, location, prep_encoding
        )

    def decode_property_style(
        self,
        prop_name: str,
        prop_schema: SchemaPath,
        location: Mapping[str, Any],
        prep_encoding: SchemaPath,
    ) -> Any:
        prop_style, prop_explode = get_style_and_explode(
            prep_encoding, default_location="query"
        )
        prop_deserializer = self.style_deserializers_factory.create(
            prop_style, prop_explode, prop_schema, name=prop_name
        )
        return prop_deserializer.deserialize(location)

    def decode_property_content_type(
        self,
        prop_name: str,
        prop_schema: SchemaPath,
        location: Mapping[str, Any],
        prop_encoding: Optional[SchemaPath] = None,
    ) -> Any:
        prop_content_type = get_content_type(prop_schema, prop_encoding)
        prop_deserializer = self.evolve(
            prop_content_type,
            prop_schema,
        )
        prop_schema_type = prop_schema.getkey("type", "")
        if (
            self.mimetype.startswith("multipart")
            and prop_schema_type == "array"
        ):
            if isinstance(location, SuportsGetAll):
                value = location.getall(prop_name)
                return list(map(prop_deserializer.deserialize, value))
            if isinstance(location, SuportsGetList):
                value = location.getlist(prop_name)
                return list(map(prop_deserializer.deserialize, value))

        return prop_deserializer.deserialize(location[prop_name])



"""OpenAPI core templating paths finders module"""

from typing import Optional

from jsonschema_path import SchemaPath
from more_itertools import peekable

from openapi_core.templating.paths.datatypes import PathOperationServer
from openapi_core.templating.paths.exceptions import OperationNotFound
from openapi_core.templating.paths.exceptions import PathNotFound
from openapi_core.templating.paths.exceptions import ServerNotFound
from openapi_core.templating.paths.iterators import SimpleOperationsIterator
from openapi_core.templating.paths.iterators import SimplePathsIterator
from openapi_core.templating.paths.iterators import SimpleServersIterator
from openapi_core.templating.paths.iterators import TemplatePathsIterator
from openapi_core.templating.paths.iterators import TemplateServersIterator
from openapi_core.templating.paths.protocols import OperationsIterator
from openapi_core.templating.paths.protocols import PathsIterator
from openapi_core.templating.paths.protocols import ServersIterator


class BasePathFinder:
    paths_iterator: PathsIterator = NotImplemented
    operations_iterator: OperationsIterator = NotImplemented
    servers_iterator: ServersIterator = NotImplemented

    def __init__(self, spec: SchemaPath, base_url: Optional[str] = None):
        self.spec = spec
        self.base_url = base_url

    def find(self, method: str, name: str) -> PathOperationServer:
        paths_iter = self.paths_iterator(
            name,
            self.spec,
            base_url=self.base_url,
        )
        paths_iter_peek = peekable(paths_iter)

        if not paths_iter_peek:
            raise PathNotFound(name)

        operations_iter = self.operations_iterator(
            method,
            paths_iter_peek,
            self.spec,
            base_url=self.base_url,
        )
        operations_iter_peek = peekable(operations_iter)

        if not operations_iter_peek:
            raise OperationNotFound(name, method)

        servers_iter = self.servers_iterator(
            name, operations_iter_peek, self.spec, base_url=self.base_url
        )

        try:
            return next(servers_iter)
        except StopIteration:
            raise ServerNotFound(name)


class APICallPathFinder(BasePathFinder):
    paths_iterator: PathsIterator = TemplatePathsIterator("paths")
    operations_iterator: OperationsIterator = SimpleOperationsIterator()
    servers_iterator: ServersIterator = TemplateServersIterator()


class WebhookPathFinder(APICallPathFinder):
    paths_iterator = SimplePathsIterator("webhooks")
    servers_iterator = SimpleServersIterator()



"""OpenAPI core validation datatypes module"""

from dataclasses import dataclass
from typing import Iterable

from openapi_core.exceptions import OpenAPIError


@dataclass
class BaseUnmarshalResult:
    errors: Iterable[OpenAPIError]

    def raise_for_errors(self) -> None:
        for error in self.errors:
            raise error



from openapi_core.contrib.aiohttp.requests import AIOHTTPOpenAPIWebRequest
from openapi_core.contrib.aiohttp.responses import AIOHTTPOpenAPIWebResponse

__all__ = [
    "AIOHTTPOpenAPIWebRequest",
    "AIOHTTPOpenAPIWebResponse",
]



from copy import deepcopy
from typing import Optional
from typing import Type

from jsonschema._format import FormatChecker
from jsonschema.protocols import Validator
from jsonschema_path import SchemaPath

from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.validators import SchemaValidator


class SchemaValidatorsFactory:
    def __init__(
        self,
        schema_validator_class: Type[Validator],
        format_checker: Optional[FormatChecker] = None,
    ):
        self.schema_validator_class = schema_validator_class
        if format_checker is None:
            format_checker = self.schema_validator_class.FORMAT_CHECKER
        self.format_checker = format_checker

    def get_format_checker(
        self,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
    ) -> FormatChecker:
        if format_validators is None:
            format_checker = deepcopy(self.format_checker)
        else:
            format_checker = FormatChecker([])
            format_checker = self._add_validators(
                format_checker, format_validators
            )
        format_checker = self._add_validators(
            format_checker, extra_format_validators
        )
        return format_checker

    def _add_validators(
        self,
        base_format_checker: FormatChecker,
        format_validators: Optional[FormatValidatorsDict] = None,
    ) -> FormatChecker:
        if format_validators is not None:
            for name, check in format_validators.items():
                base_format_checker.checks(name)(check)
        return base_format_checker

    def create(
        self,
        schema: SchemaPath,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
    ) -> Validator:
        format_checker = self.get_format_checker(
            format_validators, extra_format_validators
        )
        with schema.resolve() as resolved:
            jsonschema_validator = self.schema_validator_class(
                resolved.contents,
                _resolver=resolved.resolver,
                format_checker=format_checker,
            )

        return SchemaValidator(schema, jsonschema_validator)



"""OpenAPI core contrib werkzeug requests module"""

import re
from typing import Optional

from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict
from werkzeug.wrappers import Request

from openapi_core.datatypes import RequestParameters

# http://flask.pocoo.org/docs/1.0/quickstart/#variable-rules
PATH_PARAMETER_PATTERN = r"<(?:(?:string|int|float|path|uuid):)?(\w+)>"


class WerkzeugOpenAPIRequest:
    path_regex = re.compile(PATH_PARAMETER_PATTERN)

    def __init__(self, request: Request):
        if not isinstance(request, Request):
            raise TypeError(f"'request' argument is not type of {Request}")
        self.request = request

        self.parameters = RequestParameters(
            query=ImmutableMultiDict(self.request.args),
            header=Headers(self.request.headers),
            cookie=self.request.cookies,
        )

    @property
    def host_url(self) -> str:
        return self.request.host_url

    @property
    def path(self) -> str:
        return self.get_path(self.request.path)

    @property
    def method(self) -> str:
        return self.request.method.lower()

    @property
    def body(self) -> Optional[bytes]:
        return self.request.get_data(as_text=False)

    @property
    def content_type(self) -> str:
        # default value according to RFC 2616
        return self.request.content_type or "application/octet-stream"

    def get_path(self, path: str) -> str:
        return "".join([self.request.root_path, path])



"""OpenAPI core contrib django handlers module"""

from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Type

from django.http import JsonResponse
from django.http.request import HttpRequest
from django.http.response import HttpResponse

from openapi_core.templating.media_types.exceptions import MediaTypeNotFound
from openapi_core.templating.paths.exceptions import OperationNotFound
from openapi_core.templating.paths.exceptions import PathNotFound
from openapi_core.templating.paths.exceptions import ServerNotFound
from openapi_core.templating.security.exceptions import SecurityNotFound
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult


class DjangoOpenAPIErrorsHandler:
    OPENAPI_ERROR_STATUS: Dict[Type[BaseException], int] = {
        ServerNotFound: 400,
        SecurityNotFound: 403,
        OperationNotFound: 405,
        PathNotFound: 404,
        MediaTypeNotFound: 415,
    }

    def __call__(
        self,
        errors: Iterable[Exception],
    ) -> JsonResponse:
        data_errors = [self.format_openapi_error(err) for err in errors]
        data = {
            "errors": data_errors,
        }
        data_error_max = max(data_errors, key=self.get_error_status)
        return JsonResponse(data, status=data_error_max["status"])

    @classmethod
    def format_openapi_error(cls, error: BaseException) -> Dict[str, Any]:
        if error.__cause__ is not None:
            error = error.__cause__
        return {
            "title": str(error),
            "status": cls.OPENAPI_ERROR_STATUS.get(error.__class__, 400),
            "type": str(type(error)),
        }

    @classmethod
    def get_error_status(cls, error: Dict[str, Any]) -> str:
        return str(error["status"])


class DjangoOpenAPIValidRequestHandler:
    def __init__(self, req: HttpRequest, view: Callable[[Any], HttpResponse]):
        self.req = req
        self.view = view

    def __call__(
        self, request_unmarshal_result: RequestUnmarshalResult
    ) -> HttpResponse:
        self.req.openapi = request_unmarshal_result
        return self.view(self.req)



from dataclasses import dataclass

from openapi_core.deserializing.exceptions import DeserializeError


@dataclass
class BaseStyleDeserializeError(DeserializeError):
    """Base style deserialize operation error"""

    location: str


@dataclass
class ParameterDeserializeError(BaseStyleDeserializeError):
    """Parameter deserialize operation error"""

    style: str
    value: str

    def __str__(self) -> str:
        return (
            "Failed to deserialize value of "
            f"{self.location} parameter with style {self.style}: {self.value}"
        )


@dataclass(init=False)
class EmptyQueryParameterValue(BaseStyleDeserializeError):
    name: str

    def __init__(self, name: str):
        super().__init__(location="query")
        self.name = name

    def __str__(self) -> str:
        return (
            f"Value of {self.name} {self.location} parameter cannot be empty"
        )



"""OpenAPI core validation request module"""

from typing import Mapping

from openapi_spec_validator.versions import consts as versions
from openapi_spec_validator.versions.datatypes import SpecVersion

from openapi_core.validation.request.types import RequestValidatorType
from openapi_core.validation.request.types import WebhookRequestValidatorType
from openapi_core.validation.request.validators import V30RequestBodyValidator
from openapi_core.validation.request.validators import (
    V30RequestParametersValidator,
)
from openapi_core.validation.request.validators import (
    V30RequestSecurityValidator,
)
from openapi_core.validation.request.validators import V30RequestValidator
from openapi_core.validation.request.validators import V31RequestBodyValidator
from openapi_core.validation.request.validators import (
    V31RequestParametersValidator,
)
from openapi_core.validation.request.validators import (
    V31RequestSecurityValidator,
)
from openapi_core.validation.request.validators import V31RequestValidator
from openapi_core.validation.request.validators import (
    V31WebhookRequestBodyValidator,
)
from openapi_core.validation.request.validators import (
    V31WebhookRequestParametersValidator,
)
from openapi_core.validation.request.validators import (
    V31WebhookRequestSecurityValidator,
)
from openapi_core.validation.request.validators import (
    V31WebhookRequestValidator,
)

__all__ = [
    "VALIDATORS",
    "WEBHOOK_VALIDATORS",
    "V30RequestBodyValidator",
    "V30RequestParametersValidator",
    "V30RequestSecurityValidator",
    "V30RequestValidator",
    "V31RequestBodyValidator",
    "V31RequestParametersValidator",
    "V31RequestSecurityValidator",
    "V31RequestValidator",
    "V31WebhookRequestBodyValidator",
    "V31WebhookRequestParametersValidator",
    "V31WebhookRequestSecurityValidator",
    "V31WebhookRequestValidator",
    "V3RequestValidator",
    "V3WebhookRequestValidator",
]

# versions mapping
VALIDATORS: Mapping[SpecVersion, RequestValidatorType] = {
    versions.OPENAPIV30: V30RequestValidator,
    versions.OPENAPIV31: V31RequestValidator,
}
WEBHOOK_VALIDATORS: Mapping[SpecVersion, WebhookRequestValidatorType] = {
    versions.OPENAPIV31: V31WebhookRequestValidator,
}

# alias to the latest v3 version
V3RequestValidator = V31RequestValidator
V3WebhookRequestValidator = V31WebhookRequestValidator



"""OpenAPI core shortcuts module"""

from typing import Any
from typing import Optional
from typing import Union

from jsonschema.validators import _UNSET
from jsonschema_path import SchemaPath

from openapi_core.app import OpenAPI
from openapi_core.configurations import Config
from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.protocols import WebhookRequest
from openapi_core.types import AnyRequest
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.request.types import AnyRequestUnmarshallerType
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.request.types import (
    WebhookRequestUnmarshallerType,
)
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.unmarshalling.response.types import (
    AnyResponseUnmarshallerType,
)
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType
from openapi_core.unmarshalling.response.types import (
    WebhookResponseUnmarshallerType,
)
from openapi_core.validation.request.types import AnyRequestValidatorType
from openapi_core.validation.request.types import RequestValidatorType
from openapi_core.validation.request.types import WebhookRequestValidatorType
from openapi_core.validation.response.types import AnyResponseValidatorType
from openapi_core.validation.response.types import ResponseValidatorType
from openapi_core.validation.response.types import WebhookResponseValidatorType


def unmarshal_apicall_request(
    request: Request,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[RequestUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> RequestUnmarshalResult:
    config = Config(
        server_base_url=base_url,
        request_unmarshaller_cls=cls or _UNSET,
        **unmarshaller_kwargs,
    )
    result = OpenAPI(spec, config=config).unmarshal_apicall_request(request)
    result.raise_for_errors()
    return result


def unmarshal_webhook_request(
    request: WebhookRequest,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[WebhookRequestUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> RequestUnmarshalResult:
    config = Config(
        server_base_url=base_url,
        webhook_request_unmarshaller_cls=cls or _UNSET,
        **unmarshaller_kwargs,
    )
    result = OpenAPI(spec, config=config).unmarshal_webhook_request(request)
    result.raise_for_errors()
    return result


def unmarshal_request(
    request: AnyRequest,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[AnyRequestUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> RequestUnmarshalResult:
    config = Config(
        server_base_url=base_url,
        request_unmarshaller_cls=cls or _UNSET,
        webhook_request_unmarshaller_cls=cls or _UNSET,
        **unmarshaller_kwargs,
    )
    result = OpenAPI(spec, config=config).unmarshal_request(request)
    result.raise_for_errors()
    return result


def unmarshal_apicall_response(
    request: Request,
    response: Response,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[ResponseUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> ResponseUnmarshalResult:
    config = Config(
        server_base_url=base_url,
        response_unmarshaller_cls=cls or _UNSET,
        **unmarshaller_kwargs,
    )
    result = OpenAPI(spec, config=config).unmarshal_apicall_response(
        request, response
    )
    result.raise_for_errors()
    return result


def unmarshal_webhook_response(
    request: WebhookRequest,
    response: Response,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[WebhookResponseUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> ResponseUnmarshalResult:
    config = Config(
        server_base_url=base_url,
        webhook_response_unmarshaller_cls=cls or _UNSET,
        **unmarshaller_kwargs,
    )
    result = OpenAPI(spec, config=config).unmarshal_webhook_response(
        request, response
    )
    result.raise_for_errors()
    return result


def unmarshal_response(
    request: AnyRequest,
    response: Response,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[AnyResponseUnmarshallerType] = None,
    **unmarshaller_kwargs: Any,
) -> ResponseUnmarshalResult:
    config = Config(
        server_base_url=base_url,
        response_unmarshaller_cls=cls or _UNSET,
        webhook_response_unmarshaller_cls=cls or _UNSET,
        **unmarshaller_kwargs,
    )
    result = OpenAPI(spec, config=config).unmarshal_response(request, response)
    result.raise_for_errors()
    return result


def validate_request(
    request: AnyRequest,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[AnyRequestValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    config = Config(
        server_base_url=base_url,
        request_validator_cls=cls or _UNSET,
        webhook_request_validator_cls=cls or _UNSET,
        **validator_kwargs,
    )
    return OpenAPI(spec, config=config).validate_request(request)


def validate_response(
    request: Union[Request, WebhookRequest],
    response: Response,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[AnyResponseValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    config = Config(
        server_base_url=base_url,
        response_validator_cls=cls or _UNSET,
        webhook_response_validator_cls=cls or _UNSET,
        **validator_kwargs,
    )
    return OpenAPI(spec, config=config).validate_response(request, response)


def validate_apicall_request(
    request: Request,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[RequestValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    config = Config(
        server_base_url=base_url,
        request_validator_cls=cls or _UNSET,
        **validator_kwargs,
    )
    return OpenAPI(spec, config=config).validate_apicall_request(request)


def validate_webhook_request(
    request: WebhookRequest,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[WebhookRequestValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    config = Config(
        server_base_url=base_url,
        webhook_request_validator_cls=cls or _UNSET,
        **validator_kwargs,
    )
    return OpenAPI(spec, config=config).validate_webhook_request(request)


def validate_apicall_response(
    request: Request,
    response: Response,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[ResponseValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    config = Config(
        server_base_url=base_url,
        response_validator_cls=cls or _UNSET,
        **validator_kwargs,
    )
    return OpenAPI(spec, config=config).validate_apicall_response(
        request, response
    )


def validate_webhook_response(
    request: WebhookRequest,
    response: Response,
    spec: SchemaPath,
    base_url: Optional[str] = None,
    cls: Optional[WebhookResponseValidatorType] = None,
    **validator_kwargs: Any,
) -> None:
    config = Config(
        server_base_url=base_url,
        webhook_response_validator_cls=cls or _UNSET,
        **validator_kwargs,
    )
    return OpenAPI(spec, config=config).validate_webhook_response(
        request, response
    )



"""OpenAPI core contrib flask providers module"""

from typing import Any

from flask.globals import request
from flask.wrappers import Request


class FlaskRequestProvider:
    @classmethod
    def provide(self, *args: Any, **kwargs: Any) -> Request:
        return request



from typing import Any
from typing import Generic
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

from jsonschema_path import SchemaPath

from openapi_core.casting.schemas.exceptions import CastError
from openapi_core.schema.schemas import get_properties
from openapi_core.util import forcebool
from openapi_core.validation.schemas.validators import SchemaValidator


class PrimitiveCaster:
    def __init__(
        self,
        schema: SchemaPath,
        schema_validator: SchemaValidator,
        schema_caster: "SchemaCaster",
    ):
        self.schema = schema
        self.schema_validator = schema_validator
        self.schema_caster = schema_caster

    def __call__(self, value: Any) -> Any:
        return value


PrimitiveType = TypeVar("PrimitiveType")


class PrimitiveTypeCaster(Generic[PrimitiveType], PrimitiveCaster):
    primitive_type: Type[PrimitiveType] = NotImplemented

    def __call__(self, value: Union[str, bytes]) -> Any:
        self.validate(value)

        return self.primitive_type(value)  # type: ignore [call-arg]

    def validate(self, value: Any) -> None:
        # FIXME: don't cast data from media type deserializer
        # See https://github.com/python-openapi/openapi-core/issues/706
        # if not isinstance(value, (str, bytes)):
        #     raise ValueError("should cast only from string or bytes")
        pass


class IntegerCaster(PrimitiveTypeCaster[int]):
    primitive_type = int


class NumberCaster(PrimitiveTypeCaster[float]):
    primitive_type = float


class BooleanCaster(PrimitiveTypeCaster[bool]):
    primitive_type = bool

    def __call__(self, value: Union[str, bytes]) -> Any:
        self.validate(value)

        return self.primitive_type(forcebool(value))

    def validate(self, value: Any) -> None:
        super().validate(value)

        # FIXME: don't cast data from media type deserializer
        # See https://github.com/python-openapi/openapi-core/issues/706
        if isinstance(value, bool):
            return

        if value.lower() not in ["false", "true"]:
            raise ValueError("not a boolean format")


class ArrayCaster(PrimitiveCaster):
    @property
    def items_caster(self) -> "SchemaCaster":
        # sometimes we don't have any schema i.e. free-form objects
        items_schema = self.schema.get("items", SchemaPath.from_dict({}))
        return self.schema_caster.evolve(items_schema)

    def __call__(self, value: Any) -> List[Any]:
        # str and bytes are not arrays according to the OpenAPI spec
        if isinstance(value, (str, bytes)) or not isinstance(value, Iterable):
            raise CastError(value, self.schema["type"])

        try:
            return list(map(self.items_caster.cast, value))
        except (ValueError, TypeError):
            raise CastError(value, self.schema["type"])


class ObjectCaster(PrimitiveCaster):
    def __call__(self, value: Any) -> Any:
        return self._cast_proparties(value)

    def evolve(self, schema: SchemaPath) -> "ObjectCaster":
        cls = self.__class__

        return cls(
            schema,
            self.schema_validator.evolve(schema),
            self.schema_caster.evolve(schema),
        )

    def _cast_proparties(self, value: Any, schema_only: bool = False) -> Any:
        if not isinstance(value, dict):
            raise CastError(value, self.schema["type"])

        all_of_schemas = self.schema_validator.iter_all_of_schemas(value)
        for all_of_schema in all_of_schemas:
            all_of_properties = self.evolve(all_of_schema)._cast_proparties(
                value, schema_only=True
            )
            value.update(all_of_properties)

        for prop_name, prop_schema in get_properties(self.schema).items():
            try:
                prop_value = value[prop_name]
            except KeyError:
                continue
            value[prop_name] = self.schema_caster.evolve(prop_schema).cast(
                prop_value
            )

        if schema_only:
            return value

        additional_properties = self.schema.getkey(
            "additionalProperties", True
        )
        if additional_properties is not False:
            # free-form object
            if additional_properties is True:
                additional_prop_schema = SchemaPath.from_dict(
                    {"nullable": True}
                )
            # defined schema
            else:
                additional_prop_schema = self.schema / "additionalProperties"
            additional_prop_caster = self.schema_caster.evolve(
                additional_prop_schema
            )
            for prop_name, prop_value in value.items():
                if prop_name in value:
                    continue
                value[prop_name] = additional_prop_caster.cast(prop_value)

        return value


class TypesCaster:
    casters: Mapping[str, Type[PrimitiveCaster]] = {}
    multi: Optional[Type[PrimitiveCaster]] = None

    def __init__(
        self,
        casters: Mapping[str, Type[PrimitiveCaster]],
        default: Type[PrimitiveCaster],
        multi: Optional[Type[PrimitiveCaster]] = None,
    ):
        self.casters = casters
        self.default = default
        self.multi = multi

    def get_caster(
        self,
        schema_type: Optional[Union[Iterable[str], str]],
    ) -> Type["PrimitiveCaster"]:
        if schema_type is None:
            return self.default
        if isinstance(schema_type, Iterable) and not isinstance(
            schema_type, str
        ):
            if self.multi is None:
                raise TypeError("caster does not accept multiple types")
            return self.multi

        return self.casters[schema_type]


class SchemaCaster:
    def __init__(
        self,
        schema: SchemaPath,
        schema_validator: SchemaValidator,
        types_caster: TypesCaster,
    ):
        self.schema = schema
        self.schema_validator = schema_validator

        self.types_caster = types_caster

    def cast(self, value: Any) -> Any:
        # skip casting for nullable in OpenAPI 3.0
        if value is None and self.schema.getkey("nullable", False):
            return value

        schema_type = self.schema.getkey("type")

        type_caster = self.get_type_caster(schema_type)

        if value is None:
            return value

        try:
            return type_caster(value)
        except (ValueError, TypeError):
            raise CastError(value, schema_type)

    def get_type_caster(
        self,
        schema_type: Optional[Union[Iterable[str], str]],
    ) -> PrimitiveCaster:
        caster_cls = self.types_caster.get_caster(schema_type)
        return caster_cls(
            self.schema,
            self.schema_validator,
            self,
        )

    def evolve(self, schema: SchemaPath) -> "SchemaCaster":
        cls = self.__class__

        return cls(
            schema,
            self.schema_validator.evolve(schema),
            self.types_caster,
        )



from fastapi import Request

from openapi_core.contrib.starlette.requests import StarletteOpenAPIRequest


class FastAPIOpenAPIRequest(StarletteOpenAPIRequest):
    def __init__(self, request: Request):
        super().__init__(request)



"""OpenAPI core validation processors module"""

from openapi_core.typing import RequestType
from openapi_core.typing import ResponseType
from openapi_core.validation.integrations import ValidationIntegration


class ValidationProcessor(ValidationIntegration[RequestType, ResponseType]):
    def handle_request(self, request: RequestType) -> None:
        self.validate_request(request)

    def handle_response(
        self, request: RequestType, response: ResponseType
    ) -> None:
        self.validate_response(request, response)



from typing import Type
from typing import Union

from openapi_core.validation.request.protocols import RequestValidator
from openapi_core.validation.request.protocols import WebhookRequestValidator

RequestValidatorType = Type[RequestValidator]
WebhookRequestValidatorType = Type[WebhookRequestValidator]
AnyRequestValidatorType = Union[
    RequestValidatorType, WebhookRequestValidatorType
]



from openapi_core.contrib.werkzeug.requests import WerkzeugOpenAPIRequest
from openapi_core.contrib.werkzeug.responses import WerkzeugOpenAPIResponse

__all__ = [
    "WerkzeugOpenAPIRequest",
    "WerkzeugOpenAPIResponse",
]



"""OpenAPI core contrib requests responses module"""

from requests import Response
from werkzeug.datastructures import Headers


class RequestsOpenAPIResponse:
    def __init__(self, response: Response):
        if not isinstance(response, Response):
            raise TypeError(f"'response' argument is not type of {Response}")
        self.response = response

    @property
    def data(self) -> bytes:
        assert isinstance(self.response.content, bytes)
        return self.response.content

    @property
    def status_code(self) -> int:
        return int(self.response.status_code)

    @property
    def content_type(self) -> str:
        return str(self.response.headers.get("Content-Type", ""))

    @property
    def headers(self) -> Headers:
        return Headers(dict(self.response.headers))



"""OpenAPI core contrib falcon responses module"""

import warnings
from json import dumps
from typing import Any
from typing import Dict
from typing import Optional

from falcon.request import Request
from falcon.request import RequestOptions
from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict

from openapi_core.contrib.falcon.util import unpack_params
from openapi_core.datatypes import RequestParameters


class FalconOpenAPIRequest:
    def __init__(
        self,
        request: Request,
        default_when_empty: Optional[Dict[Any, Any]] = None,
    ):
        if not isinstance(request, Request):
            raise TypeError(f"'request' argument is not type of {Request}")
        self.request = request
        if default_when_empty is None:
            default_when_empty = {}
        self.default_when_empty = default_when_empty

        # Path gets deduced by path finder against spec
        self.parameters = RequestParameters(
            query=ImmutableMultiDict(unpack_params(self.request.params)),
            header=Headers(self.request.headers),
            cookie=self.request.cookies,
        )

    @property
    def host_url(self) -> str:
        assert isinstance(self.request.prefix, str)
        return self.request.prefix

    @property
    def path(self) -> str:
        assert isinstance(self.request.path, str)
        return self.request.path

    @property
    def method(self) -> str:
        assert isinstance(self.request.method, str)
        return self.request.method.lower()

    @property
    def body(self) -> Optional[bytes]:
        # Falcon doesn't store raw request stream.
        # That's why we need to revert deserialized data

        # Support falcon-jsonify.
        if hasattr(self.request, "json"):
            return dumps(self.request.json).encode("utf-8")

        media = self.request.get_media(
            default_when_empty=self.default_when_empty,
        )
        handler, _, _ = self.request.options.media_handlers._resolve(
            self.request.content_type, self.request.options.default_media_type
        )
        try:
            body = handler.serialize(
                media, content_type=self.request.content_type
            )
        # multipart form serialization is not supported
        except NotImplementedError:
            warnings.warn(
                f"body serialization for {self.request.content_type} not supported"
            )
            return None
        else:
            assert isinstance(body, bytes)
            return body

    @property
    def content_type(self) -> str:
        if self.request.content_type:
            assert isinstance(self.request.content_type, str)
            return self.request.content_type

        assert isinstance(self.request.options, RequestOptions)
        assert isinstance(self.request.options.default_media_type, str)
        return self.request.options.default_media_type



"""OpenAPI core contrib starlette handlers module"""

from typing import Any
from typing import Dict
from typing import Iterable
from typing import Type

from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.responses import Response

from openapi_core.templating.media_types.exceptions import MediaTypeNotFound
from openapi_core.templating.paths.exceptions import OperationNotFound
from openapi_core.templating.paths.exceptions import PathNotFound
from openapi_core.templating.paths.exceptions import ServerNotFound
from openapi_core.templating.security.exceptions import SecurityNotFound
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult


class StarletteOpenAPIErrorsHandler:
    OPENAPI_ERROR_STATUS: Dict[Type[BaseException], int] = {
        ServerNotFound: 400,
        SecurityNotFound: 403,
        OperationNotFound: 405,
        PathNotFound: 404,
        MediaTypeNotFound: 415,
    }

    def __call__(
        self,
        errors: Iterable[Exception],
    ) -> JSONResponse:
        data_errors = [self.format_openapi_error(err) for err in errors]
        data = {
            "errors": data_errors,
        }
        data_error_max = max(data_errors, key=self.get_error_status)
        return JSONResponse(data, status_code=data_error_max["status"])

    @classmethod
    def format_openapi_error(cls, error: BaseException) -> Dict[str, Any]:
        if error.__cause__ is not None:
            error = error.__cause__
        return {
            "title": str(error),
            "status": cls.OPENAPI_ERROR_STATUS.get(error.__class__, 400),
            "type": str(type(error)),
        }

    @classmethod
    def get_error_status(cls, error: Dict[str, Any]) -> str:
        return str(error["status"])


class StarletteOpenAPIValidRequestHandler:
    def __init__(self, request: Request, call_next: RequestResponseEndpoint):
        self.request = request
        self.call_next = call_next

    async def __call__(
        self, request_unmarshal_result: RequestUnmarshalResult
    ) -> Response:
        self.request.scope["openapi"] = request_unmarshal_result
        return await self.call_next(self.request)
